/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */
package uk.ac.roslin.ensembl.config;

import java.io.PrintWriter;
import java.io.Serializable;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Locale;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.logging.Level;
import org.apache.log4j.Logger;
import uk.ac.roslin.ensembl.exception.ConfigurationException;

public class DBConnection {

    protected String newline = (System.getProperty("line.separator") != null) ? System.getProperty("line.separator") : "\r\n";
    String datasourceStyle = "ensembldb";
    String driver = null;
    String url = null;
    String username = "anonymous";
    String password = null;
    String[] invalidDBMatches = null;
    String[] invalidDBContains = null;
    String[] invalidDBStart = null;
    String[] invalidDBEnd = null;
    DataSource configuredDatasource;
    Properties dbProperties;
    final static Logger LOGGER = Logger.getLogger(DBConnection.class);

    public static enum DataSource implements Serializable {

        ENSEMBLDB("EnsemblDB"),
        ENSEMBLDB_ARCHIVES("EnsemblDB Archives"),
        ENSEMBLGENOMES("EnsemblGenomes"),
        LOCAL("Local Data Source");
        private String label;

        DataSource(String label) {
            this.label = label;
        }

        @Override
        public String toString() {
            return label;
        }
    }

    public DBConnection(DataSource datasource) throws ConfigurationException {
        configuredDatasource = datasource;
        if (datasource == DataSource.ENSEMBLDB) {
            dbProperties = this.readResource("uk.ac.roslin.ensembl.configfiles.ensembldb");
        } else if (datasource == DataSource.ENSEMBLDB_ARCHIVES) {
            dbProperties = this.readResource("uk.ac.roslin.ensembl.configfiles.ensembldb_archives");
        } else if (datasource == DataSource.ENSEMBLGENOMES) {
            dbProperties = this.readResource("uk.ac.roslin.ensembl.configfiles.ensemblgenomes");
        } else {
            throw new ConfigurationException("Unrecognized type of DataSource specified for DBConnection");
        }
        initialize();
    }

    public Properties getConfigurationProperties() {
        return dbProperties;
    }

    public DBConnection(DataSource datasource, Properties local_datasource) throws ConfigurationException {
        if (datasource != DataSource.LOCAL || local_datasource == null) {
            throw new ConfigurationException("Invalid attempt to configure DBConnection with unrecognized local DataSource.");
        }
        configuredDatasource = datasource;
        dbProperties = local_datasource;
        initialize();
    }

    private Properties readResource(String id) throws ConfigurationException {
        Properties p = null;
        try {
            //NB: need to pass in the classloader to get this to work in a test environment!
            ResourceBundle rb = ResourceBundle.getBundle(id, Locale.getDefault(),
                this.getClass().getClassLoader());
            p = new Properties();
            for (Enumeration keys = rb.getKeys(); keys.hasMoreElements();) {
                final String key = (String) keys.nextElement();
                final String value = rb.getString(key);
                p.put(key, value);
            }
        } catch (Exception ex) {
            throw new ConfigurationException("System can't read the configuration file: " + id);
        }
        return p;
    }

    private void initialize() {


        if (dbProperties.getProperty("invalidDBStartsWith") != null
                && !dbProperties.getProperty("invalidDBStartsWith").trim().isEmpty()) {
            invalidDBStart = dbProperties.getProperty("invalidDBStartsWith").split(" ");
        }
        if (dbProperties.getProperty("invalidDBEndsWith") != null
                && !dbProperties.getProperty("invalidDBEndsWith").trim().isEmpty()) {
            invalidDBEnd = dbProperties.getProperty("invalidDBEndsWith").split(" ");
        }
        if (dbProperties.getProperty("invalidDBContains") != null
                && !dbProperties.getProperty("invalidDBContains").trim().isEmpty()) {
            invalidDBContains = dbProperties.getProperty("invalidDBContains").split(" ");
        }
        if (dbProperties.getProperty("invalidDBMatches") != null
                && !dbProperties.getProperty("invalidDBMatches").trim().isEmpty()) {
            invalidDBMatches = dbProperties.getProperty("invalidDBMatches").split(" ");
        }

        if (LOGGER.isInfoEnabled()) {
            if (this.invalidDBContains != null) {
                LOGGER.info("skipping databases containing String: " + Arrays.toString(this.invalidDBContains));
            }
            if (this.invalidDBMatches != null) {
                LOGGER.info("skipping databases matching String: " + Arrays.toString(this.invalidDBMatches));
            }
            if (this.invalidDBStart != null) {
                LOGGER.info("skipping databases beginning with String: " + Arrays.toString(this.invalidDBStart));
            }
            if (this.invalidDBEnd != null) {
                LOGGER.info("skipping databases ending with String: " + Arrays.toString(this.invalidDBEnd));
            }
        }

        if (dbProperties.getProperty("datasource") != null
                && !dbProperties.getProperty("datasource").trim().isEmpty()) {
            datasourceStyle = dbProperties.getProperty("datasource").trim();
        }
        if (dbProperties.getProperty("url") != null
                && !dbProperties.getProperty("url").trim().isEmpty()) {
            url = dbProperties.getProperty("url").trim();
        }
        if (dbProperties.getProperty("driver") != null
                && !dbProperties.getProperty("driver").trim().isEmpty()) {
            driver = dbProperties.getProperty("driver").trim();
        }
        if (dbProperties.getProperty("username") != null
                && !dbProperties.getProperty("username").trim().isEmpty()) {
            username = dbProperties.getProperty("username").trim();
        }
        if (dbProperties.getProperty("password") != null
                && !dbProperties.getProperty("password").trim().isEmpty()) {
            password = dbProperties.getProperty("password").trim();
        }

        if (LOGGER.isInfoEnabled()) {
            LOGGER.info("datasource: " + datasourceStyle);
            LOGGER.info("driver: " + driver);
            LOGGER.info("url: " + url);
            LOGGER.info("username: " + username);
            LOGGER.info("password: " + password);
        }

    }

    public Boolean isDBNameValid(String s) {
        if (this.invalidDBContains != null && this.invalidDBContains.length > 0) {
            for (int i = 0; i < this.invalidDBContains.length; i++) {
                if (s.contains(this.invalidDBContains[i])) {
                    return false;
                }
            }
        }
        if (this.invalidDBStart != null && this.invalidDBStart.length > 0) {
            for (int i = 0; i < this.invalidDBStart.length; i++) {
                if (s.startsWith(this.invalidDBStart[i])) {
                    return false;
                }
            }
        }
        if (this.invalidDBEnd != null && this.invalidDBEnd.length > 0) {
            for (int i = 0; i < this.invalidDBEnd.length; i++) {
                if (s.endsWith(this.invalidDBEnd[i])) {
                    return false;
                }
            }
        }
        if (this.invalidDBMatches != null && this.invalidDBMatches.length > 0) {
            for (int i = 0; i < this.invalidDBMatches.length; i++) {
                if (s.equalsIgnoreCase(this.invalidDBMatches[i])) {
                    return false;
                }
            }
        }
        return true;
    }

    public String getDatasourceStyle() {
        return datasourceStyle;
    }

    public String report() {
        String out = "DB Connection Report" + newline + "--------------"+newline+newline;

        out = out + "Datasource Configured: " + configuredDatasource.toString() + newline;
        out = out + "Datasource style: " + datasourceStyle + newline;
        out = out + "Datasource URL: " + url + newline;
        out = out + "Datasource user: " + username + newline;

        if (this.invalidDBContains != null) {
            out = out.concat("Skipping databases containing String: " + Arrays.toString(this.invalidDBContains)+newline);
        }
        if (this.invalidDBMatches != null) {
            out = out.concat("Skipping databases matching String: " + Arrays.toString(this.invalidDBMatches)+newline);
        }
        if (this.invalidDBStart != null) {
            out = out.concat("Skipping databases beginning with String: " + Arrays.toString(this.invalidDBStart)+newline);
        }
        if (this.invalidDBEnd != null) {
            out = out.concat("Skipping databases ending with String: " + Arrays.toString(this.invalidDBEnd)+newline);
        }

        out = out+newline;
        return out;
    }

    public String testConnection() {
        String warning = "";

        try {
            Class.forName("com.mysql.jdbc.Driver");
            PrintWriter w = new PrintWriter(System.err);
            DriverManager.setLogWriter(w);
            Connection conn = DriverManager.getConnection(url, username, password);

            SQLWarning warn = conn.getWarnings();

            if (warn != null) {
                warning = "WARNING:\n";
            }

            while (warn != null) {
                warning += "SQLState: " + warn.getSQLState() + "\n";
                warning += "Message:  " + warn.getMessage() + "\n";
                warning += "Vendor:   " + warn.getErrorCode() + "\n";
                warn = warn.getNextWarning();
            }


            conn.close();

        } catch (ClassNotFoundException e) {
            warning += "ERROR: Can't load driver\n" + e;
        } catch (SQLException e) {
            warning += "ERROR: Database access failed\n" + e;
    }
        if (warning.isEmpty()) {
            return "OK";
        } else {
            return warning;
        }

    }

    public boolean isConnectionOK() {

        try {
            Class.forName("com.mysql.jdbc.Driver");
            PrintWriter w = new PrintWriter(System.err);
            DriverManager.setLogWriter(w);
            Connection conn = DriverManager.getConnection(url, username, password);

            SQLWarning warn = conn.getWarnings();

            if (warn != null) {
                return false;
            }
            conn.close();
        } catch (Exception e) {
            return false;
        }

        return true;

    }



}
