/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */
package uk.ac.roslin.ensembl.config;

import java.util.Collection;
import java.util.HashMap;
import java.util.Locale;
import uk.ac.roslin.ensembl.model.database.DatabaseType;

/**
 * In 'ensembl.org' there is a single 'ensembl_compara' database which contains
 * the pairwise comparison data on all the species in 'ensembl.org. However, in
 * 'ensemblgenomes.org', there are separate 'ensembl_compara' databases for the
 * different taxonomic groupings:
 * <ul><li>ensembl_compara_plants</li>
 * <li>ensembl_compara_bacteria</li>
 * <li>ensembl_compara_fungi</li>
 * <li>ensembl_compara_metazoa</li>
 * <li>ensembl_compara_protozoa</li>
 * <li>ensembl_compara_pan_homology</li></ul>
 * In ensembl.org all species are compared in the ensembl_compara database, and
 * belong to the compara group 'multi'.  In each core database (species) in
 * ensemblgenomes.org the 'species.division'  key in table 'meta' specifies the
 * group:
 * <ul><li>EnsemblPlants</li>
 * <li>EnsemblBacteria</li>
 * <li>EnsemblFungi</li>
 * <li>EnsemblMetazoa</li>
 * <li>EnsemblProtozoa</li></ul>
 * Orthology / paralogy (peptide) predictions are done between all species of a database,
 * but genomic alignments only between certain combinations of species. Bacteria
 * is a slightly special case in that we compare members of each collection to
 * each other, but not between the collections (its actually 10 comparas merged
 * into one).
 * <p>ensembl_compara_pan_homology is a peptide compara database produced from
 * a set of selected species that are taken from all EnsemblGenomes divisions and
 * from Ensembl (but doesn't include all species from all divisions). See list at
 * <a href="http://metazoa.ensembl.org/info/docs/compara/homology_method.html">
 * http://metazoa.ensembl.org/info/docs/compara/homology_method.html</a>.
 * One thing to note here is that pan also contains protein family data as well:
 * <a href="http://metazoa.ensembl.org/info/docs/compara/family.html">
 * http://metazoa.ensembl.org/info/docs/compara/family.html</a>. This means that
 * the genome_db table ends up with more entries than are used in the peptide
 * homology comparisons alone and cannot be used to determine membership, so you
 * need to use the species_set/method_link tables as well rather than just relying
 * on genome_db.
 */
public class EnsemblComparaDivision extends EnsemblType implements DatabaseType, Comparable<EnsemblComparaDivision> {

    public static EnsemblComparaDivision MULTI;
    public static EnsemblComparaDivision PLANTS;
    public static EnsemblComparaDivision PROTISTS;
    public static EnsemblComparaDivision BACTERIA;
    public static EnsemblComparaDivision FUNGI;
    public static EnsemblComparaDivision METAZOA;
    public static EnsemblComparaDivision PAN_HOMOLOGY;

    private EnsemblComparaDivision(String value) {
        this.label = value;
    }

    public static String getDBName(String label) {
        String out = "ensembl_compara";
        if (label.equalsIgnoreCase("multi")) {
            return out;
        } else {
            return out+"_"+label;
        }
    }

    private static HashMap<String, EnsemblComparaDivision> typeListHash = EnsemblComparaDivision.initialize();

    public static Collection<EnsemblComparaDivision> getAllTypes() {
        return typeListHash.values();
    }

    private static HashMap<String, EnsemblComparaDivision> initialize() {

        HashMap<String, EnsemblComparaDivision> out = new HashMap<String, EnsemblComparaDivision>();

        MULTI = new EnsemblComparaDivision("multi");
        PLANTS = new EnsemblComparaDivision("plants");
        PROTISTS = new EnsemblComparaDivision("protists");
        BACTERIA = new EnsemblComparaDivision("bacteria");
        FUNGI = new EnsemblComparaDivision("fungi");
        METAZOA = new EnsemblComparaDivision("metazoa");
        PAN_HOMOLOGY = new EnsemblComparaDivision("pan_homology");


        out.put(MULTI.toString(),MULTI );
        out.put(PLANTS.toString(), PLANTS);
        out.put(PROTISTS.toString(),PROTISTS );
        out.put(BACTERIA.toString(), BACTERIA );
        out.put(FUNGI.toString(),FUNGI );
        out.put(METAZOA.toString(),METAZOA );
        out.put(PAN_HOMOLOGY.toString(), PAN_HOMOLOGY );

        return out;
    }

    public static EnsemblComparaDivision getEnsemblComparaDivision(String value) {
        return typeListHash.get(value);
    }

    public static EnsemblComparaDivision getEnsemblComparaDivisionByMetaValue(String value) {
        String v = null;
        if (value != null) {
            v = value.replace("Ensembl","").toLowerCase();
        }
        return typeListHash.get(v);
    }



    public int compareTo(EnsemblComparaDivision other) {
        if (other == null || other.toString() == null) {
            return -1;
        }

        return this.toString().compareTo(other.toString());
    }
}
