/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */

package uk.ac.roslin.ensembl.config;

import java.io.ObjectStreamException;
import java.io.Serializable;
import java.util.Collection;
import java.util.HashMap;
import java.util.TreeSet;
import uk.ac.roslin.ensembl.model.database.DatabaseType;

/**
 * Concrete implementation of DatabaseType interface for representing static
 * Ensembl DatabaseTypes. Implements static strings representing each of the current enesmbl
 * database types, and concrete DatabaseTypes for each of these string types.
 * @author paterson
 */
public class EnsemblDBType extends EnsemblType implements DatabaseType, Comparable<EnsemblDBType> {

     /**
     * The SpecificDBType of a given EnsemblDBType Object
     * i.e. CORE, COMPARA etc.
     */
    private SpecificDBType databaseType;

    /**
     * The GenericDBType of a given EnsemblDBType Object
     * i.e. SINGLE_SPECIES, SPECIES_COMPARISON etc.
     */
   private GenericDBType ENSEMBL_TYPE ;

    /**
     * Static set holding the Database Types use in cross species comparisons
     */
    private static TreeSet<EnsemblDBType> speciesComparisonDBTypes = new TreeSet<EnsemblDBType>();

    /**
     * Static set holding the single species Database Types
     */
    private static TreeSet<EnsemblDBType> singleSpeciesDBTypes = new TreeSet<EnsemblDBType>();

    /**
     * Static set holding the multi-species Database Types in ensembl genomes
     */
    private static TreeSet<EnsemblDBType> collectionDBTypes = new TreeSet<EnsemblDBType>();
       
    /**
     * Enumeration of the GenericDBTypes 
     */
    private static enum GenericDBType implements Serializable {
        SPECIES_COMPARISON("species_comparison"),
        SINGLE_SPECIES("single_species"),
        COLLECTION("collection");
        
        private String label;
        
        GenericDBType (String label){
            this.label=label;
        }

        @Override
        public String toString() {
            return label;
        }
    }

     /**
     * Enumeration of the SpecificDBTypes
     */
    private static enum SpecificDBType  implements Serializable {
        
        CDNA("cdna"),
        CORE("core"),
        COREEXPRESSIONEST("coreexpressionest"),
        COREEXPRESSIONGNF("coreexpressiongnf"),
        FUNCGEN("funcgen"),
        OTHERFEATURES("otherfeatures"),
        VARIATION("variation"),
        VEGA("vega"),
        ANCESTRAL("ensembl_ancestral"),
        COMPARA("ensembl_compara"),
        GO("ensembl_go"),
        ONTOLOGY("ensembl_ontology"),
        WEBSITE("ensembl_website"),
        COLLECTION_CORE("collection_core"),
        COLLECTION_FUNCGEN("collection_funcgen"),
        COLLECTION_VARIATION("collection_variation");    
        
        private String label;
        
        SpecificDBType (String label){
            this.label=label;
        }

        @Override
        public String toString() {
            return label;
        }        
        
        
    }
            
    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl species cDNA database sType
     */
    public static  EnsemblDBType cdna;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl species core database sType
     */
    public static EnsemblDBType core;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl species coreexpressioncdna database sType
     */
    public static EnsemblDBType coreexpressionest;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl species coreexpressiongnf database sType
     */
    public static EnsemblDBType coreexpressiongnf;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl species funcgen database sType
     */
    public static EnsemblDBType funcgen;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl species otherfeatures database sType
     */
    public static EnsemblDBType otherfeatures;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl species variation database sType
     */
    public static EnsemblDBType variation;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl species vega database sType
     */
    public static EnsemblDBType vega;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl ancestral database sType
     */
    public static EnsemblDBType ancestral;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl compara database sType
     */
    public static EnsemblDBType compara;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl go database sType
     */
    public static EnsemblDBType go;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl ontology database sType
     */
    public static EnsemblDBType ontology;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensembl website database sType
     */
    public static EnsemblDBType website;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensemblgenomes collection core database sType
     */
    public static EnsemblDBType collection_core;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensemblgenomes collection funcgen database sType
     */
    public static EnsemblDBType collection_funcgen;

    /**
     * A concrete instantiation of an DatabaseType object representing the ensemblgenomes collection variation database sType
     */
    public static EnsemblDBType collection_variation;

    /**
     *
     */
    private static HashMap<String, EnsemblDBType> typeHashList = EnsemblDBType.init();

     /**
     * A private accessor to make a concrete instantiation of a EnsemblDBType object for a given string name.
     * This implements a singleton pattern against each name.
     * @param typeName  String The name of the EnsemblDBType object to be generated.
     * @return EnsemblDBType the object
     */
    private static EnsemblDBType makeDatabaseType(SpecificDBType sType, GenericDBType gType) {

        //we have to initialize the hash first time called
        if (EnsemblDBType.typeHashList == null) {
            EnsemblDBType.typeHashList = new HashMap<String, EnsemblDBType>();
        }

        //if we already have this type use it
        EnsemblDBType dt = getDBTypeForName(sType.toString());
        //if not make it, and put it in the relevant hashes
        if (dt == null) {
            dt = new EnsemblDBType(sType, gType);
            EnsemblDBType.typeHashList.put(sType.toString(), dt);

            if (dt.ENSEMBL_TYPE.equals(GenericDBType.SPECIES_COMPARISON)) {
                speciesComparisonDBTypes.add(dt);
            }
            else if (dt.ENSEMBL_TYPE.equals(GenericDBType.SINGLE_SPECIES)) {
                singleSpeciesDBTypes.add(dt);
            }
            else if (dt.ENSEMBL_TYPE.equals(GenericDBType.COLLECTION)) {
                collectionDBTypes.add(dt);
            }
        }
        return dt;
    }

     /**
     * A private constructor for making EnsemblDBType objects. Should only be
     * accessed from within the class - and then only the first time that
     * the class is accessed when the list of acceptable types is constructed.
     * @param sType String The name/sType of this object
     */
    private EnsemblDBType(SpecificDBType sType, GenericDBType gType) {
        this.databaseType = sType;
        this.ENSEMBL_TYPE = gType;
    }

    /**
     * private method to make all the EnsemblDBTypes and put them in a hash
     * @return HashMap the initialized and populated hash of all the types
     */
    private static HashMap init() {
        cdna = EnsemblDBType.makeDatabaseType(SpecificDBType.CDNA, GenericDBType.SINGLE_SPECIES);
        core = EnsemblDBType.makeDatabaseType(SpecificDBType.CORE, GenericDBType.SINGLE_SPECIES);
        coreexpressionest = EnsemblDBType.makeDatabaseType(SpecificDBType.COREEXPRESSIONEST, GenericDBType.SINGLE_SPECIES);
        coreexpressiongnf = EnsemblDBType.makeDatabaseType(SpecificDBType.COREEXPRESSIONGNF, GenericDBType.SINGLE_SPECIES);
        funcgen = EnsemblDBType.makeDatabaseType(SpecificDBType.FUNCGEN, GenericDBType.SINGLE_SPECIES);
        otherfeatures = EnsemblDBType.makeDatabaseType(SpecificDBType.OTHERFEATURES, GenericDBType.SINGLE_SPECIES);
        variation = EnsemblDBType.makeDatabaseType(SpecificDBType.VARIATION, GenericDBType.SINGLE_SPECIES);
        vega = EnsemblDBType.makeDatabaseType(SpecificDBType.VEGA, GenericDBType.SINGLE_SPECIES);
        ancestral = EnsemblDBType.makeDatabaseType(SpecificDBType.ANCESTRAL, GenericDBType.SPECIES_COMPARISON);
        compara = EnsemblDBType.makeDatabaseType(SpecificDBType.COMPARA, GenericDBType.SPECIES_COMPARISON);
        go = EnsemblDBType.makeDatabaseType(SpecificDBType.GO, GenericDBType.SPECIES_COMPARISON);
        ontology = EnsemblDBType.makeDatabaseType(SpecificDBType.ONTOLOGY, GenericDBType.SPECIES_COMPARISON);
        website = EnsemblDBType.makeDatabaseType(SpecificDBType.WEBSITE, GenericDBType.SPECIES_COMPARISON);
        collection_core = EnsemblDBType.makeDatabaseType(SpecificDBType.COLLECTION_CORE, GenericDBType.COLLECTION);
        collection_funcgen = EnsemblDBType.makeDatabaseType(SpecificDBType.COLLECTION_FUNCGEN, GenericDBType.COLLECTION);
        collection_variation = EnsemblDBType.makeDatabaseType(SpecificDBType.COLLECTION_VARIATION, GenericDBType.COLLECTION);

        return typeHashList;

    }

    /**
     * Ensure Singleton class
     */
    private Object readResolve() throws ObjectStreamException {
        return this;
    }

    /**
     * We often use the constants in a String context, so this routine allows
     * us to do that easily
     * @return The String name of this EnsemblDBType object
     */
    @Override
    public String toString() {
        return this.databaseType.toString();
    }

    /**
     * Overrides the Object method to test for equality
     * @param   obj   the reference object with which to compare.
     * @return  <code>true</code> if this object is the same as the obj
     *          argument; <code>false</code> otherwise.
     */
    @Override
    public boolean equals(Object obj) {
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }

        EnsemblDBType other = (EnsemblDBType)obj;
        String otherString = other.toString();
        return this.toString().equals(otherString);

    }

    /**
     * Overrides Object method to generate a hashCode
     * @return  a hash code value for this object.
     */
    @Override
    public int hashCode() {
        int hash = 7;
        hash = 31 * hash + (this.toString() != null ? this.toString().hashCode() : 0);
        return hash;

    }

    /**
     * Public Method to return the static EnsemblDBType for a typeName.
     * Checks to see if the supplied sType is the name of a valid EnsemblDBType
     * and returns the corresponding EnsemblDBType object if it is, or returns null
     * if not; also returns null if the  typeHashList is not initialized.
     * @param typeName String
     * @return EnsemblDBType for the query typeName
     */
    public static EnsemblDBType getDBTypeForName (String typeName) {
        EnsemblDBType dt = null;
        if ( EnsemblDBType.typeHashList != null
                && EnsemblDBType.typeHashList.containsKey(typeName)) {
            dt =  (EnsemblDBType) EnsemblDBType.typeHashList.get(typeName);
        }
        return dt;
    }

    public static TreeSet<EnsemblDBType> getSpeciesComparisonDatabaseTypes() {
        return EnsemblDBType.speciesComparisonDBTypes;

    }

    public static TreeSet<EnsemblDBType> getSingleSpeciesDatabaseTypes() {
        return EnsemblDBType.singleSpeciesDBTypes;

    }

    public static TreeSet<EnsemblDBType> getCollectionDatabaseTypes() {
        return EnsemblDBType.collectionDBTypes;

    }

    public static Collection<EnsemblDBType> getAllDatabaseTypes() {
        return EnsemblDBType.typeHashList.values();
    }

    @Override
    public int compareTo(EnsemblDBType other) {
        if (other == null || other.toString() == null) {
            return -1;
        }
            return this.toString().compareTo(other.toString());
    }



}
