/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */
package uk.ac.roslin.ensembl.config;

import java.io.File;
import java.io.FileReader;
import java.util.Properties;
import org.apache.log4j.Logger;
import uk.ac.roslin.ensembl.config.DBConnection.DataSource;
import uk.ac.roslin.ensembl.exception.ConfigurationException;

public class RegistryConfiguration {

    private DataSource type;
    private DBConnection db = null;
    private SchemaVersion schema = null;

    final static Logger LOGGER = Logger.getLogger(RegistryConfiguration.class);

    /**
     * Constructor for a  non-configured RegistryConfiguration object. This is not a valid
     * Configuration until a DBConnection object is set.
     */
    public RegistryConfiguration() {
    }

    /**
     * Constructs a RegistryConfiguration object after calling setType, which will
     * set the DBConnection if DataSource is of a known valid type. If no 
     * further setting of schemaProperties this object will return 
     * default configuration values read from properties files in ensembl-config jar. 
     * @param type DataSource
     * @throws ConfigurationException
     */
    public RegistryConfiguration(DataSource type) throws ConfigurationException {
        this.setType(type);
    }

    /**
     * Sets the DBConnection on this RegistryConfiguration object
     * if DataSource is of a known valid type.
     * @param type
     * @throws ConfigurationException
     */
    private void setType(DataSource type) throws ConfigurationException {
        this.type = type;
        if (this.type == DataSource.ENSEMBLDB) {
            db = new DBConnection(DataSource.ENSEMBLDB);
        } else if (this.type == DataSource.ENSEMBLDB_ARCHIVES) {
            db = new DBConnection(DataSource.ENSEMBLDB_ARCHIVES);
        } else if (this.type == DataSource.ENSEMBLGENOMES) {
            db = new DBConnection(DataSource.ENSEMBLGENOMES);
        }
    }

    /**
     * Sets DBConnection on this RegistryConfiguration object
     * from a given set of Properties.
     * @param dbProperties Properties
     * @throws ConfigurationException
     */
    public void setDBProperties(Properties dbProperties) throws ConfigurationException {
        this.type = DataSource.LOCAL;
        db = new DBConnection(type, dbProperties);
    }

    /**
     * Sets the SchemaVersion object on this RegistryConfiguration object
     * using the given Properties.
     * @param schemaProperties Properties
     * @throws ConfigurationException
     */
    public void setSchemaProperties(Properties schemaProperties) throws ConfigurationException {
        schema = new SchemaVersion(schemaProperties);
    }

    /**
     * Sets DBConnection on this RegistryConfiguration object
     * from a given File of Properties.
     * @param dbProperties File
     * @throws ConfigurationException
     */
    public void setDBByFile(File dbProperties) throws ConfigurationException {
        this.type = DataSource.LOCAL;
        if (dbProperties == null || !dbProperties.canRead()) {
            throw new ConfigurationException("Failure to read properties from local DB configuration file");
        }

        Properties props = new Properties();
        try {
            props.load(new FileReader(dbProperties));
            if (props == null || props.isEmpty()) {
                throw new Exception();
            }
            db = new DBConnection(type, props);
        } catch (Exception ex) {
            throw new ConfigurationException("Failure to read properties from local DB configuration file", ex);
        }
    }

    /**
     * Sets the SchemaVersion object on this RegistryConfiguration object
     * using a File with Properties values.
     * @param schemaMappings File
     * @throws ConfigurationException
     */
    public void setSchemaByFile(File schemaMappings) throws ConfigurationException {

        if (schemaMappings == null || !schemaMappings.canRead()) {
            throw new ConfigurationException("Failure to read schema mappings from local configuration file");
        }

        Properties props = new Properties();
        try {
            props.load(new FileReader(schemaMappings));
            if (props == null || props.isEmpty()) {
                throw new Exception();
            }
            schema = new SchemaVersion(props);
        } catch (Exception ex) {
            throw new ConfigurationException("Failure to read properties from local mapping rules configuration file", ex);
        }
    }

    /**
     * Returns the DBCOnnection object set  on this RegistryConfigurationobject,
     * or throws an ConfigurationException if there is no DBConnection set.
     * @return DBConnection
     * @throws ConfigurationException
     */
    public DBConnection getDb() throws ConfigurationException {
        if (db==null) {
             throw new ConfigurationException("No DB connection properties set");
        }
        return db;
    }


    /**
     * Returns the SchemaVersionobject set on this RegistryConfiguration object.
     * If SchemaVersion is unset, it is first initialized to a default version
     * (read from properties file in ensembl-config artifact).
     * @return SchemaVersion
     * @throws ConfigurationException
     */
    public SchemaVersion getSchema() throws ConfigurationException {
        if (schema==null) {
            schema = new SchemaVersion();
        }
        if (schema==null) {
            throw new ConfigurationException("No schema mapping rules set");
        }
        return schema;
    }

    /**
     * Returns the current DataSource type of this RegistryConfiguration object.
     * @return DataSource
     */
    public DataSource getType() {
        return type;
    }



}
