/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */
package uk.ac.roslin.ensembl.config;

import java.util.Properties;
import java.util.Enumeration;
import java.util.Locale;
import java.util.ResourceBundle;
import org.apache.log4j.Logger;
import uk.ac.roslin.ensembl.exception.ConfigurationException;

/**
 *
 * @author paterson
 */
public class SchemaVersion {

    final static Logger LOGGER = Logger.getLogger(SchemaVersion.class);

    private Properties schemaVersionProperties = null;
    private String currentEnsemblVersion = null;
    private String currentGenomesVersion = null;
    private String[] registeredSchemaVersions = null;

    public SchemaVersion() throws ConfigurationException {
        schemaVersionProperties = this.readResource("uk.ac.roslin.ensembl.configfiles.schema_version_mappings");
        initialize();
    }

    public SchemaVersion(Properties localMappingRules) throws ConfigurationException {
        schemaVersionProperties = localMappingRules;
        if (schemaVersionProperties == null || schemaVersionProperties.isEmpty()) {
            throw new ConfigurationException("No Schema Mapping properties given");
        }
        initialize();
    }

    private void initialize() {
        currentEnsemblVersion = (schemaVersionProperties != null) ? schemaVersionProperties.getProperty("current_ensembl_release") : null;
        currentGenomesVersion = (schemaVersionProperties != null) ? schemaVersionProperties.getProperty("current_genomes_release") : null;
        registeredSchemaVersions = (schemaVersionProperties != null
                && schemaVersionProperties.getProperty("known_ensembl_schemas") != null)
                ? schemaVersionProperties.getProperty("known_ensembl_schemas").split(" ")
                : null;
    }

    private Properties readResource(String id) throws ConfigurationException {
        Properties p = null;
        try {
            //NB: need to pass in the classloader to get this to work in a test environment!
            ResourceBundle rb = ResourceBundle.getBundle(id, Locale.getDefault(),
                this.getClass().getClassLoader());
            p = new Properties();
            for (Enumeration keys = rb.getKeys(); keys.hasMoreElements();) {
                final String key = (String) keys.nextElement();
                final String value = rb.getString(key);
                p.put(key, value);
            }

        } catch (Exception ex) {
            throw new ConfigurationException("System can't read the configuration file: " + id, ex);
        }

        if (p == null || p.isEmpty()) {
            throw new ConfigurationException("No Schema Mapping properties read from " + id);
        }
        return p;
    }

    public String getMybatisSchemaPath(String database, String version) {
        String out = null;
        String location = null;
        String sc = null;

        location = (schemaVersionProperties != null) ? schemaVersionProperties.getProperty("schema_location") : null;
        sc = (schemaVersionProperties != null) ? schemaVersionProperties.getProperty(database + "_" + version + "_schema") : null;

        out = (location != null && sc != null) ? location + sc + "Configuration.xml" : null;
        return out;
    }

    public Properties getConfigurationProperties() {
        return schemaVersionProperties;
    }

    public String getCurrentEnsemblVersion() {
        return currentEnsemblVersion;

    }

    public String getCurrentGenomesVersion() {
        return currentGenomesVersion;
    }

    public String[] getRegisteredSchemas() {
        return registeredSchemaVersions;
    }

    public String getBaseMybatis() {
        return schemaVersionProperties.getProperty("base_mybatis");
    }
}
