/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */

package uk.ac.roslin.ensembl.dao.database;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeMap;
import uk.ac.roslin.ensembl.datasourceaware.core.DAChromosome;
import uk.ac.roslin.ensembl.datasourceaware.core.DACoordinateSystem;
import uk.ac.roslin.ensembl.dao.database.factory.DBDAOCollectionCoreFactory;
import uk.ac.roslin.ensembl.dao.database.factory.DBDAOCollectionFactory;
import uk.ac.roslin.ensembl.exception.ConfigurationException;
import uk.ac.roslin.ensembl.exception.DAOException;
import uk.ac.roslin.ensembl.model.ObjectType;
import uk.ac.roslin.ensembl.model.core.CoordinateSystem;
import uk.ac.roslin.ensembl.model.core.Species;
import uk.ac.roslin.ensembl.model.database.CollectionCoreDatabase;
import uk.ac.roslin.ensembl.model.core.CollectionSpecies;
import uk.ac.roslin.ensembl.config.EnsemblCoordSystemType;
import uk.ac.roslin.ensembl.config.EnsemblDBType;
import uk.ac.roslin.ensembl.config.FeatureType;
import uk.ac.roslin.ensembl.model.database.Registry;

/**
 *
 * @author paterson
 */
public class DBCollectionCoreDatabase extends DBCollectionDatabase  implements CollectionCoreDatabase {


    private TreeMap<DBCollectionSpecies, TreeMap<Integer, DACoordinateSystem>> CSHash = new TreeMap<DBCollectionSpecies, TreeMap<Integer, DACoordinateSystem>>();

    private TreeMap<DBCollectionSpecies, DACoordinateSystem> seqLevelCSHash = new TreeMap<DBCollectionSpecies, DACoordinateSystem>();
    private TreeMap<DBCollectionSpecies, DACoordinateSystem> chrLevelCSHash = new TreeMap<DBCollectionSpecies, DACoordinateSystem>();
    private TreeMap<DBCollectionSpecies, DACoordinateSystem> topLevelCSHash = new TreeMap<DBCollectionSpecies, DACoordinateSystem>();

    private TreeMap<DBCollectionSpecies, DBDAOCollectionCoreFactory> factoryHash = new TreeMap<DBCollectionSpecies, DBDAOCollectionCoreFactory>();

    private TreeMap<DBCollectionSpecies, HashMap<FeatureType, HashMap<DACoordinateSystem, Integer>>>  featureCSHash
            = new TreeMap<DBCollectionSpecies, HashMap<FeatureType, HashMap<DACoordinateSystem, Integer>>>();

    private TreeMap<DBCollectionSpecies, HashMap<DACoordinateSystem, List<FeatureType>>>    CSFeatureHash
            = new TreeMap<DBCollectionSpecies, HashMap<DACoordinateSystem, List<FeatureType>>>();


//    private TreeMap<DBCollectionSpecies, String> speciesGeneBuildLevelHash = new TreeMap<DBCollectionSpecies, String>();
//    private TreeMap<DBCollectionSpecies, String> speciesExonBuildLevelHash = new TreeMap<DBCollectionSpecies, String>();
//    private TreeMap<DBCollectionSpecies, String> speciesTranscriptBuildLevelHash = new TreeMap<DBCollectionSpecies, String>();

    private TreeMap<DBCollectionSpecies, HashMap<FeatureType, String>> speciesBuildLevels = new TreeMap<DBCollectionSpecies, HashMap<FeatureType,String>>();

    public DBCollectionCoreDatabase(String db_name, EnsemblDBType type, Registry registry) throws ConfigurationException {
        super(db_name, type, registry);
    }


    @Override
    public DBDAOCollectionCoreFactory getCoreFactory(Species sp) {



        if (sp == null || collection ==null || !this.collection.getSpecies().contains(sp)) {
            return null;
        }

        DBCollectionSpecies species = (DBCollectionSpecies) sp;

        if (this.factoryHash.containsKey(species)) {
            return this.factoryHash.get(species);
        } else {
            DBDAOCollectionCoreFactory f = null;

            try {
                f = (DBDAOCollectionCoreFactory) DBDAOCollectionFactory.makeFactory(this,  species);
                this.factoryHash.put(species, f);
            } catch (Exception ex) {
            }
            return f;

        }
    }

    @Override
    public DACoordinateSystem getChromosomeLevelCS(Species species) throws DAOException {

        if (species == null ||  this.collection == null || !this.getCollection().getSpecies().contains(species)) {
            return null;
        }

        DBCollectionSpecies sp = (DBCollectionSpecies) species;

        if (!this.chrLevelCSHash.containsKey(sp)) {
            this.lazyLoadCoordinateSytems(sp);
        }

        return this.chrLevelCSHash.get(sp);

    }

    @Override
    public DACoordinateSystem getSequenceLevelCS(Species species) throws DAOException {
        
        if (species == null || this.collection == null || !this.getCollection().getSpecies().contains(species)) {
            return null;
        }
        DBCollectionSpecies sp = (DBCollectionSpecies) species;
        if (!this.seqLevelCSHash.containsKey(sp)) {
            this.lazyLoadCoordinateSytems(sp);
        }

        return this.seqLevelCSHash.get(sp);
    }

    @Override
    public DACoordinateSystem getTopLevelCS(Species species) throws DAOException {
        
        if ( species == null || this.collection == null || !this.getCollection().getSpecies().contains(species)) {
            return null;
        }
        DBCollectionSpecies sp = (DBCollectionSpecies) species;
        if (!this.topLevelCSHash.containsKey(sp)) {
            this.lazyLoadCoordinateSytems(sp);
        }

        return this.topLevelCSHash.get(sp);
    }


    private List<DACoordinateSystem> fetchCS(DBCollectionSpecies species) throws DAOException {

        List<DACoordinateSystem> coords = new ArrayList<DACoordinateSystem>();
        if (species == null || collection ==null || !this.collection.getSpecies().contains(species)) {
            return coords;
        }

        //can throw DAOException
        if ( this.getCoreFactory(species)!=null) {
            coords = this.getCoreFactory(species).getCoordinateSystemDAO().
                getCoordinateSystems();
        } else {
           return coords;
        }
        return coords;
    }


    private void lazyLoadCoordinateSytems(DBCollectionSpecies species)  throws DAOException {

        if (species == null || collection ==null || !this.collection.getSpecies().contains(species)) {
            return ;
        }

        List<DACoordinateSystem> coords = this.fetchCS(species);

        if (!coords.isEmpty()) {
               this.setCoordinateSytems(species, coords);
        }

            //set null values to prevent repeating lazy load

            if (!this.chrLevelCSHash.containsKey(species)) {
                this.chrLevelCSHash.put(species, null);
            }
            if (!this.topLevelCSHash.containsKey(species)) {
                this.topLevelCSHash.put(species, null);
            }
            if (!this.seqLevelCSHash.containsKey(species)) {
                this.seqLevelCSHash.put(species, null);
            }

    }

    private void setCoordinateSytems(DBCollectionSpecies species, List<DACoordinateSystem> coords ) throws DAOException {

        if (species == null || collection ==null || !this.collection.getSpecies().contains(species)) {
            return ;
        }

        TreeMap<Integer, DACoordinateSystem> hash1 = new TreeMap<Integer, DACoordinateSystem>();


        //set  the is for the top, chromosome and sequence levels

        for (DACoordinateSystem cs: coords) {

            hash1.put(cs.getId(), cs);

            if (cs.getRank()==1) {

                this.topLevelCSHash.put(species, cs);

                if (cs.getType()==EnsemblCoordSystemType.chromosome) {
                    this.chrLevelCSHash.put(species, cs);
                }
            }

            if (cs.isSequenceLevel()) {
                    this.seqLevelCSHash.put(species, cs);
                }
        }

        this.CSHash.put(species, hash1);

        //if the top rank wasnt chromosome level, find the highest ranked chromosome level

        if (!this.chrLevelCSHash.containsKey(species)) {

            TreeMap <Integer, DACoordinateSystem> temp = new TreeMap <Integer, DACoordinateSystem>();

            for (DACoordinateSystem cs: coords) {
                temp.put(cs.getRank(), cs);
            }

            Iterator<Integer> it = temp.keySet().iterator();
            while (it.hasNext()) {
                DACoordinateSystem cs = temp.get(it.next());
                if (cs.getType()==EnsemblCoordSystemType.chromosome) {
                    this.chrLevelCSHash.put(species, cs);
                    break;
                }
            }
        }
        }

    @Override
    public void setBuildLevels(Species species, HashMap<String, String> keyedValues) {

        if (species == null || collection ==null || !this.collection.getSpecies().contains(species)) {
            return ;
        }

        HashMap<FeatureType, String> map = new HashMap<FeatureType, String>();

        for (String key : keyedValues.keySet() ) {

            String temp = key.replace("build.level", "");

            if (FeatureType.getFeatureType(temp)!= null ) {
                map.put(FeatureType.getFeatureType(temp), keyedValues.get(key));
            }

        }

        this.speciesBuildLevels.put((DBCollectionSpecies) species, map);
    }

    @Override
    @SuppressWarnings("element-type-mismatch")
    public String getBuildLevel(Species species, String featureKey) throws DAOException{

        if (species == null || collection ==null || !this.collection.getSpecies().contains(species)) {
            return null;
        }

        String out = null;

        if (this.speciesBuildLevels.get(species)!= null) {
            out = this.speciesBuildLevels.get(species).get(FeatureType.getFeatureType(featureKey));
        }
        return out;
    }

    //dont really want this public - just for testing
    public HashMap<FeatureType, String> getBuildLevels(CollectionSpecies species) {
        if (species == null || collection ==null || !this.collection.getSpecies().contains(species)) {
            return null;
        }
        return this.speciesBuildLevels.get(species);
    }

    private void setFeatureCSHash(DBCollectionSpecies sp) throws DAOException {

        if (sp == null || collection ==null || !this.collection.getSpecies().contains(sp)) {
            return;
        }

        if (!this.CSHash.containsKey(sp)) {
            this.lazyLoadCoordinateSytems(sp);
        }


        HashMap<DACoordinateSystem, List<FeatureType>> hash1 = new HashMap<DACoordinateSystem, List<FeatureType>>();

        for (DACoordinateSystem c : this.CSHash.get(sp).values()) {
            hash1.put(c, new ArrayList<FeatureType>());

        }

        this.CSFeatureHash.put(sp, hash1);

        this.featureCSHash.put(sp,  new  HashMap<FeatureType, HashMap<DACoordinateSystem, Integer>>() );

        try {

            this.getCoreFactory(sp).getCoordinateSystemDAO().setFeatureCS();

        } catch (DAOException e) {
            throw e;
        }

    }

    @Override
    public void addFeatureCS(String featureType, Integer csID, Integer maxLength, Species sp) {

        if (sp == null || collection ==null || !this.collection.getSpecies().contains(sp)) {
            return ;
        }

        FeatureType localType = null;
        DACoordinateSystem cs = null;

        localType = FeatureType.getFeatureType(featureType);
        cs = this.CSHash.get(sp).get(csID);

        if (localType == null || cs == null) {
            return;
        } else {
            if (!this.featureCSHash.get(sp).containsKey(localType)) {
                this.featureCSHash.get(sp).put(localType, new HashMap<DACoordinateSystem, Integer>());
            }
            this.featureCSHash.get(sp).get(localType).put(cs, maxLength);

            this.CSFeatureHash.get(sp).get(cs).add(localType);

        }


    }


    @Override
    public Set<DACoordinateSystem> getCSForFeature(Species sp, ObjectType feature) throws DAOException {

        if (sp == null || collection ==null || !this.collection.getSpecies().contains(sp)) {
            return null;
        }

            if (!this.featureCSHash.containsKey(sp)) {
                this.setFeatureCSHash((DBCollectionSpecies) sp);
            }


        if (this.featureCSHash.containsKey(sp)
                && this.featureCSHash.get(sp).containsKey(feature)) {
            return this.featureCSHash.get(sp).get(feature).keySet();
        } else {

            return null;
        }
    }

    @Override
    public List<FeatureType> getFeaturesForCS(Species sp, CoordinateSystem coordSys) throws DAOException {

        if (sp== null || collection ==null || !this.collection.getSpecies().contains(sp)) {
            return null;
        }

            if (!this.featureCSHash.containsKey(sp)) {
                this.setFeatureCSHash((DBCollectionSpecies) sp);
            }

        if (this.CSFeatureHash.containsKey(sp)
                && this.CSFeatureHash.get(sp).containsKey(coordSys)) {
            return this.CSFeatureHash.get(sp).get(coordSys);
        } else {
            return null;
        }
    }

    @Override
    public Integer getMaxLengthForFeature(Species sp, ObjectType feature, CoordinateSystem cs) throws DAOException {

        if (sp == null || collection ==null || !this.collection.getSpecies().contains(sp)) {
            return null;
        }

        if ( feature == null || cs == null ) {
            return null;
        }

            if (!this.featureCSHash.containsKey(sp) ) {
                this.setFeatureCSHash((DBCollectionSpecies) sp);
            }

        if (this.featureCSHash.containsKey(sp)
                && this.featureCSHash.get(sp).containsKey(feature)
                && this.featureCSHash.get(sp).get(feature).containsKey(cs)) {
            return this.featureCSHash.get(sp).get(feature).get(cs); 
        } else {
            return null;
        }

    }

    @Override
    public DACoordinateSystem getBuildCoordSystem(Species species, String featureType) throws DAOException{

        if (species == null || collection ==null || !this.collection.getSpecies().contains(species)
                || this.speciesBuildLevels.get(species) == null ||
                featureType ==null || featureType.isEmpty()) {
            return null;
        }

            if (FeatureType.getFeatureType(featureType) != null) {
                String level = this.speciesBuildLevels.get(species).get(FeatureType.getFeatureType(featureType));
                if (level.equalsIgnoreCase("toplevel")) {
                    return this.topLevelCSHash.get(species);
                } else {
                    return null;
                }

//                //are there other values possible?
//                else if (level.equalsIgnoreCase("chromosomelevel")) {
//                    return this.getChromosomeLevelCoordSystem();
//                } else if (level.equalsIgnoreCase("sequencelevel")) {
//                    return this.getSequenceLevelCoordSystem();
//                }
            }

        return null;
    }

    @Override
    public DACoordinateSystem getCSByID(Species species, Integer id) throws DAOException{

        if (species == null || collection ==null || !this.collection.getSpecies().contains(species)
                || id == null) {
            return null;
        }

        DACoordinateSystem out = null;
        if (!this.CSHash.containsKey(species)) {

                //lazyload can throw DAOException
                this.lazyLoadCoordinateSytems((DBCollectionSpecies) species);

        }

        out = this.CSHash.get(species).get(id); 

        return out;
    }

    @Override
    public DAChromosome getChromosomeByName(Species species, String name) throws DAOException {

        if (species == null || collection ==null || !this.collection.getSpecies().contains(species)) {
            return null;
        }

        return this.getCoreFactory(species).getChromosomeDAO().getChromosomeByName(name);

    }

    @Override
    public List<DAChromosome> getChromosomes(Species species) throws DAOException {

        if (species == null || collection ==null || !this.collection.getSpecies().contains(species)) {
            return null;
        }

        return this.getCoreFactory(species).getChromosomeDAO().getChromosomes();

    }


}
