/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (dbVersion 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */
package uk.ac.roslin.ensembl.dao.database;


import java.io.InputStreamReader;
import java.io.Reader;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.logging.Level;
import uk.ac.roslin.ensembl.config.DBConnection.DataSource;
import uk.ac.roslin.ensembl.exception.DAOException;
import java.util.Properties;
import java.util.TreeMap;
import java.util.TreeSet;
import org.apache.ibatis.session.SqlSession;
import org.apache.ibatis.session.SqlSessionFactory;
import org.apache.ibatis.session.SqlSessionFactoryBuilder;
import org.apache.log4j.Logger;
import uk.ac.roslin.ensembl.config.DBConnection;
import uk.ac.roslin.ensembl.model.database.CollectionDatabase;
import uk.ac.roslin.ensembl.model.database.Database;
import uk.ac.roslin.ensembl.model.database.DatabaseType;
import uk.ac.roslin.ensembl.exception.ConfigurationException;
import uk.ac.roslin.ensembl.model.core.Species;
import uk.ac.roslin.ensembl.model.database.CollectionCoreDatabase;
import uk.ac.roslin.ensembl.model.database.Registry;
import uk.ac.roslin.ensembl.config.EnsemblDBType;
import uk.ac.roslin.ensembl.config.EnsemblComparaDivision;
import uk.ac.roslin.ensembl.config.FeatureType;
import uk.ac.roslin.ensembl.config.RegistryConfiguration;
import uk.ac.roslin.ensembl.config.SchemaVersion;
import uk.ac.roslin.ensembl.mapper.DatabaseMapper;
import uk.ac.roslin.ensembl.model.core.CollectionSpecies;
import uk.ac.roslin.ensembl.model.database.ComparisonDatabase;
import uk.ac.roslin.ensembl.model.database.SingleSpeciesCoreDatabase;
import uk.ac.roslin.ensembl.model.database.SingleSpeciesDatabase;

public class DBRegistry
        implements Registry {

    final static Logger LOGGER = Logger.getLogger(DBRegistry.class);

    protected SqlSessionFactory sqlMapper;

    protected Properties configProperties = new Properties();
    protected SchemaVersion schemaVersion = null;
    protected DBConnection dbConnection = null;
    private String mostRecentEnsemblVersion;
    private String mostRecentGenomesVersion;

    //incase the datasource does not include the 'current' dbVersion
    private int highestEnsemblVersion = 0;
    private int highestVersion = 0;

    protected TreeSet<Integer> knownSchemaVersions = new TreeSet<Integer>();

    //the mybatis configProperties for database connection
    private String baseMybatis;

    //to read the mybatis regConfig
    protected Reader reader;

    protected TreeMap<String, DBSpecies> speciesHash = new TreeMap<String, DBSpecies>();
    protected TreeMap<String, DBCollectionSpecies> collectionspeciesHash = new TreeMap<String, DBCollectionSpecies>();
    protected List<DBCollection> collections = new ArrayList<DBCollection>();

    //we need a separate one of these for each collection (dbVersion) as they will grow
    protected TreeMap<String, TreeMap<String, DBCollection>> collectionNameDBVersionHash = new TreeMap<String, TreeMap<String, DBCollection>>();
    protected TreeMap<String, String> ensemblNameForAliasHash = new TreeMap<String, String>();

    //the databases
    protected TreeSet<DBDatabase> allDatabases = new TreeSet<DBDatabase>();
    protected TreeSet<DBSingleSpeciesDatabase> singleSpeciesDatabases = new TreeSet<DBSingleSpeciesDatabase>();
    protected TreeSet<DBComparisonDatabase> comparisonDatabases = new TreeSet<DBComparisonDatabase>();
    
    //HashMap of comparison databases indexed on 'group' , then on 'dbVersion' (not schema dbVersion)
    protected HashMap<EnsemblComparaDivision, HashMap<String,DBComparisonDatabase>> comparaDBByDivision =
            new HashMap<EnsemblComparaDivision, HashMap<String,DBComparisonDatabase>>();

    protected TreeSet<DBCollectionDatabase> collectionDatabases = new TreeSet<DBCollectionDatabase>();
    protected String newline = (System.getProperty("line.separator") != null) ? System.getProperty("line.separator") : "\r\n";
    protected StringBuffer tooNewDB = new StringBuffer();
    protected StringBuffer unknownDB = new StringBuffer();

    protected RegistryConfiguration regConfig = null;
    DataSource datasourceType;
    protected boolean alreadyInitialized = false;

    /**
     * Parameterless public constructor for DBRegistry object
     */
    public DBRegistry()  {
        
    }

    /**
     * Public Constructor for default value DBRegistry object, using values in
     * ensembl-config artifact. This intializes the database connection so will fail
     * if connection parameters are wrong etc.
     * @param type DBConnection.DataSource
     * @throws ConfigurationException
     * @throws DAOException
     */
    public DBRegistry(DataSource type) throws ConfigurationException, DAOException {
        this.setConfiguration(type);
        this.initialize();
    }
    /**
     * Public Constructor for default value DBRegistry object, using values in
     * ensembl-config artifact. Boolean parameter allows a REgistry to be configured
     * without calling initialization (useful for tests);
     * @param type DBConnection.DataSource
     * @param initialize boolean
     * @throws ConfigurationException
     * @throws DAOException
     */
    public DBRegistry(RegistryConfiguration conf, boolean initialize) throws ConfigurationException, DAOException {
        this.setConfiguration(conf);
        if (initialize) {
             this.initialize();
        }

    }

    public void setConfiguration(RegistryConfiguration conf) throws ConfigurationException, DAOException {
        this.regConfig = conf;
        this.datasourceType = regConfig.getType();
        this.dbConnection = regConfig.getDb();
        this.configProperties.putAll(dbConnection.getConfigurationProperties());
        this.schemaVersion = regConfig.getSchema();
        this.configProperties.putAll(schemaVersion.getConfigurationProperties());
        this.mostRecentEnsemblVersion = this.schemaVersion.getCurrentEnsemblVersion();
        this.mostRecentGenomesVersion = this.schemaVersion.getCurrentGenomesVersion();
        String[] versions = schemaVersion.getRegisteredSchemas();

        for (int i = 0; i < versions.length; i++) {
            this.knownSchemaVersions.add(Integer.parseInt(versions[i]));
        }

        LOGGER.info("This application is configured to use schema version " + mostRecentEnsemblVersion
                + " of Ensembl, and knows about the schema versions: " + Arrays.toString(versions));


        // mybatis configProperties for databaseDAO
        this.baseMybatis = this.schemaVersion.getBaseMybatis();
        try {
            reader = new InputStreamReader(this.getClass().getClassLoader().getResourceAsStream(baseMybatis));
            //this doesn't work once you bundle up
            //reader = Resources.getResourceAsReader(baseMybatis);
        } catch (Exception ex) {
            throw new ConfigurationException("Fail to read Mybatis Configuration for initial Database connection", ex);
        }
        try {
            sqlMapper = new SqlSessionFactoryBuilder().build(reader, "current", configProperties);
            if (sqlMapper==null) {
                throw new Exception("SqlSessionFactory is null");
            }
        } catch (Exception e) {
            throw new ConfigurationException("The DBRegistry is unable to make a SqlSessionFactory", e);
        }
    }

    public void setConfiguration(DataSource type) throws ConfigurationException, DAOException {
        this.setConfiguration(new RegistryConfiguration(type));
    }

//    /**
//     * Public Constructor for a DBRegistry, to be configures and initialized using
//     * settings in the RegistryConfiguration parameter.
//     * @param conf RegistryConfiguration
//     * @throws ConfigurationException
//     * @throws DAOException
//     */
//    public DBRegistry(RegistryConfiguration conf) throws ConfigurationException, DAOException {
//        this.setConfiguration(conf);
//    }

    /**
     * private initialisation method for the registry which connects to ensembl 
     * via a DBDatasourceDAO and retreives details on all the current tables and 
     * creates database objects of the appropriate type for each.
     * Also retrieves dbSpeciesName aliases for all the current release dbSpeciesName.
     * And finds out what dbSpeciesName are present in multispecies databases.
     *
     * @throws DAOException, ConfigurationException
     */
    public void initialize() throws DAOException, ConfigurationException {

        if (this.alreadyInitialized) {
            throw new ConfigurationException("Tried to re-initialize a DBRegistry");
        } 

        //populates the names of all the databases available, excluding marts, test, mysql and information_schema
        LOGGER.info("Registry is loading and sorting all databases available from this source ("
                + this.dbConnection.getConfigurationProperties().getProperty("url") + "):");

        for (String s : this.getDatabases()) {

            if (dbConnection.isDBNameValid(s)) {
                try {
                    DBDatabase db = this.makeDatabase(s.trim());

                    if (db.intSchemaVersion > Integer.parseInt(this.mostRecentEnsemblVersion)) {
                        LOGGER.warn("IGNORING Database " + s + " with higher schema number ("
                                + db.getSchemaVersion() + ") than application configuration (" + this.mostRecentEnsemblVersion + ").");
                        tooNewDB.append(s + newline);
                    } else if (!this.knownSchemaVersions.contains(Integer.parseInt(db.getSchemaVersion()))) {
                        LOGGER.warn("IGNORING Database " + s + " with schema number ("
                                + db.getSchemaVersion() + ") not found in the application configuration.");
                        unknownDB.append(s + newline);
                    } else {
                        this.registerDatabase(db);
                    }

                } catch (Exception e) {
                    LOGGER.warn("Unspecified error thrown trying to make a Database object for '" + s + "'", e);
                }
            }
        }



        //get all the core databases for the current release
        //so that we can create Species database objects for each of these
        //filling in the aliases and some other properties
        //we then put these in a hash against genus_species

        //BUG: if the datasource doesnt have any current releases
        //e.g. if configured to load ensembldb_archives
        //we wont make any species -
        //need a work around if this is the case - to use the highest release dbVersion found


        TreeSet<DBSingleSpeciesDatabase> badDatabases = new TreeSet<DBSingleSpeciesDatabase>();

        for (DBSingleSpeciesDatabase d : this.singleSpeciesDatabases) {

            if (d.type == EnsemblDBType.core && d.getIntDBVersion() == this.highestVersion) {
                
                DBSpecies spData = null;

                //this can fail if there is something wrong with the database!
                try {
                    spData = this.getSpeciesProperties((DBSingleSpeciesCoreDatabase) d);
                } catch (DAOException dAOE) {
                    LOGGER.info("Failed to create a species for current core database: "+d.getdBName()
                            +": Removed this database!", dAOE );
                }
                if (spData != null) {
                    d.setSpecies(spData);
                    speciesHash.put(spData.getDatabaseStyleName(), spData);
                } else {
                    //store this as a bad (current) database
                    badDatabases.add(d);
                }
            }

        }

        //remove any databases that i cant make a species for!
        this.singleSpeciesDatabases.removeAll(badDatabases);
        this.allDatabases.removeAll(badDatabases);



        // the collections will add members with time - so each collectionDB needs
        //its own Collection

        //get all the collection_core databases 
        //so that we can create Collection database objects for each of these
        //and populate it with CollectionSpecies objects for this collection

        // however we will reuse existing CollectionSpecies objects
        // and only add new ones to the registry


       TreeSet<DBCollectionDatabase> badCollectionDatabases = new TreeSet<DBCollectionDatabase>();

        for (DBCollectionDatabase d : this.collectionDatabases) {
            if (d.type == EnsemblDBType.collection_core) {

                // make a registrycollection for each collection core database
                //as these may differ over time
                DBCollection thisCollection = null;
                try {
                    thisCollection = this.getCollectionSpeciesProperties((CollectionCoreDatabase) d);
                } catch (DAOException ex) {
                    LOGGER.info("Failed to create a collectionfor current core database: "+d.getdBName()
                            +": Removed this database!", ex);
                }

                if (thisCollection!=null) {

                    this.collections.add(thisCollection);
                    //d.setCollection(r);

                    if (this.collectionNameDBVersionHash.containsKey(thisCollection.getCollectionName())) {
                        this.collectionNameDBVersionHash.get(thisCollection.getCollectionName()).put(d.getDBVersion(), thisCollection);
                    } else {
                        TreeMap<String, DBCollection> map = new TreeMap<String, DBCollection>();
                        map.put(d.getDBVersion(), thisCollection);
                        this.collectionNameDBVersionHash.put(thisCollection.getCollectionName(), map);
                    }
                } else {
                    badCollectionDatabases.add(d);
                }
            }
        }

        this.collectionDatabases.removeAll(badCollectionDatabases);
        this.allDatabases.removeAll(badCollectionDatabases);

        // we need to set the collections on the other types of collection databases
        for (DBCollectionDatabase d : this.collectionDatabases) {
            if (d.type != EnsemblDBType.collection_core) {

                String dbVersion = d.getDBVersion();
                String cName = d.getCollectionName();

                if (this.collectionNameDBVersionHash.containsKey(cName)
                        && this.collectionNameDBVersionHash.get(cName).containsKey(dbVersion)) {
                    d.setCollection(this.collectionNameDBVersionHash.get(cName).get(dbVersion));
                }

            }
        }



        //we now populate the look up hash for aliases
        //and at the same time add all the relevant databases to each DBSpecies
        //and vice versa

        for (DBSpecies sp : speciesHash.values()) {


            for (String s : sp.getAliases()) {
                this.ensemblNameForAliasHash.put(s, sp.getDatabaseStyleName());
            }

            for (DBSingleSpeciesDatabase d : this.singleSpeciesDatabases) {
                if (d.getDbSpeciesName().equalsIgnoreCase(sp.getDatabaseStyleName())) {
                    sp.addDatabase(d);
                    if (d.getSpecies() == null) {
                        d.setSpecies(sp);
                    }
                }
            }

        }

        for (DBCollectionSpecies sp : collectionspeciesHash.values()) {
            for (String s : sp.getAliases()) {
                this.ensemblNameForAliasHash.put(s, sp.getDatabaseStyleName());
            }


            for (DBCollectionDatabase d : this.collectionDatabases) {
                try {

                    if (EnsemblDBType.getCollectionDatabaseTypes().contains(d.getType())
                            && d.getCollection().getSpecies().contains(sp)) {
                        sp.addDatabase(d);
                    }
                } catch (Exception e) {
                }

            }



        }

        //add all the relevant databases to each DBCollection


        Iterator<String> it = this.collectionNameDBVersionHash.keySet().iterator();

        while (it.hasNext()) {

            String name = it.next();

            Iterator<String> it2 = this.collectionNameDBVersionHash.get(name).keySet().iterator();

            while (it2.hasNext()) {
                String dbversion = it2.next();

                DBCollection collection = this.collectionNameDBVersionHash.get(name).get(dbversion);

                TreeSet<CollectionDatabase> temp = new TreeSet<CollectionDatabase>();
                for (DBCollectionDatabase d : this.collectionDatabases) {
                    if (d.getCollectionName().equals(name) && d.getDBVersion().equalsIgnoreCase(dbversion.toString())) {
                        temp.add((CollectionDatabase) d);
                    }
                }
                collection.addDatabases(temp);

            }
        }
        this.alreadyInitialized = true;
    }

    // Database look up methods
    @Override
    public DBDatabase getDatabase(String species_name_or_alias) {
        return getDatabase(species_name_or_alias, null, null);
    }

    @Override
    public DBDatabase getDatabase(String species_name_or_alias, String db_version) {
        return getDatabase(species_name_or_alias, null, db_version);
    }

    @Override
    public DBDatabase getDatabase(String speciesNameOrAlias_collectionName_comparaDivision, DatabaseType database_type, String db_version) {

        String name = speciesNameOrAlias_collectionName_comparaDivision;

        DBDatabase out = null;

        //default to CORE type if not provided
        EnsemblDBType type = (database_type != null
                && database_type.toString() != null
                && !database_type.toString().equals(""))
                ? (EnsemblDBType) database_type
                : EnsemblDBType.core;

       if (name==null || name.isEmpty()) {
            if (EnsemblDBType.getSpeciesComparisonDatabaseTypes().contains(type)) {
                name = "multi";
            } else {
                //deff out if no name
                return out;
            }
        }

        //default to current dbVersion if not provided
        String version = (db_version != null
                && !db_version.isEmpty()
                && !db_version.equals("current"))
                ? db_version
                : this.getMostRecentDBVersion();

        String speciesOrGroup = "";

        if (EnsemblDBType.getSpeciesComparisonDatabaseTypes().contains(type)) {
            //no look up possible
            speciesOrGroup = name;
        } else if (EnsemblDBType.getSingleSpeciesDatabaseTypes().contains(type)) {
            //do a look up for aliases
            speciesOrGroup = this.getEnsemblNameForAlias(name);
            if (speciesOrGroup == null) {
                //deff out
                return out;
            }
        } else if (EnsemblDBType.getCollectionDatabaseTypes().contains(type)) {
            //no look up possible
            speciesOrGroup = name;
        }


        if (EnsemblDBType.getSingleSpeciesDatabaseTypes().contains(type)) {
            for (DBSingleSpeciesDatabase d : this.singleSpeciesDatabases) {
                if (d.getDbSpeciesName().equals(speciesOrGroup)
                        && d.getDBVersion().equals(version)
                        && d.getType() == type) {
                    return d;
                }
            }
        } else if (EnsemblDBType.getSpeciesComparisonDatabaseTypes().contains(type)) {

            EnsemblComparaDivision gp = EnsemblComparaDivision.getEnsemblComparaDivision(speciesOrGroup);
            for (DBComparisonDatabase d : this.comparisonDatabases) {
                if (d.getComparisonDivision().equals(gp)
                        && d.getDBVersion().equals(version)
                        && d.getType() == type) {
                    return d;
                }
            }
        }        //this is only working on Groups not ind species
        else if (EnsemblDBType.getCollectionDatabaseTypes().contains(type)) {
            for (DBCollectionDatabase d : this.collectionDatabases) {
                if (d.getCollectionName().equals(speciesOrGroup)
                        && d.getDBVersion().equals(version)
                        && d.getType() == type) {
                    return d;
                }
            }
        }


        return out;
    }

    @Override
    public ComparisonDatabase getComparaDatabase(EnsemblComparaDivision comparaDivision, String db_version) {

        if (this.comparaDBByDivision.get(comparaDivision) != null) {
            return this.comparaDBByDivision.get(comparaDivision).get(db_version);
        } else {
            return null;
        }
    }

    @Override
    public DBDatabase getDatabaseForFullName(String fullname) {

        DBDatabase out = null;

        if (fullname==null || fullname.isEmpty()) {
            return out;
        }

        for (DBDatabase d : this.allDatabases) {
            if (d.getdBName().equals(fullname)) {
                out = d;
                break;
            }
        }
        return out;
    }

    @Override
    public TreeSet<DBDatabase> getDatabasesByType(DatabaseType type) {
        TreeSet<DBDatabase> out = new TreeSet<DBDatabase>();

        for (DBDatabase d : this.allDatabases) {
            if (d.getType() == type) {
                out.add(d);
            }
        }
        return out;
    }

    @Override
    public String findMybatisSchemaForSchemaVersion(DatabaseType type, String schema_version) {

        String out = null;

        try {
            out = schemaVersion.getMybatisSchemaPath(type.toString(), schema_version);
        } catch (Exception e) {
        }

        //

        if (out == null) {
            try {
                out = schemaVersion.getMybatisSchemaPath(type.toString(), schemaVersion.getCurrentEnsemblVersion());
            } catch (Exception e) {
            }
        }

        return out;
    }

    /**
     * Looks up the ensembl_genus_species_name for the given alias from the
     * ensemblNameForAliasHash hash of aliases in this Registry.
     * @param alias String
     * @return String the ensembl_genus_species_name
     */
    @Override
    public String getEnsemblNameForAlias(String alias) {
        if (alias==null) {
            return null;
        }
        if (ensemblNameForAliasHash.containsValue(alias)) {
            return alias;
        }
        return ensemblNameForAliasHash.get(alias);
    }

    /**
     * Method for an adhoc alias to be added by the user in this session. Adds
     * the alias to the aliases TreeSet of the cognate DBSpecies and to
     * the ensemblNameForAliasHash hash of aliases in this Registry
     * @param ensembl_genus_species_name Ensmebl databse dbVersion of the name, e.g. 'homo_sapiens'
     * @param alias any string the user wants to use an alias for this
     * @return boolean true if successful
     * @throws Exception if update fails, with meaning full message
     */
    @Override
    public boolean addSessionAlias(String ensembl_genus_species_name, String alias) throws Exception {

        if(ensembl_genus_species_name==null || alias==null || ensembl_genus_species_name.isEmpty() || alias.isEmpty() ) {
            throw new Exception("Failed to add alias: " + alias + " for species: " + ensembl_genus_species_name);
        }

        Species spData = null;


        try {
            spData = this.speciesHash.get(ensembl_genus_species_name);
            if (spData == null) {
                spData = this.collectionspeciesHash.get(ensembl_genus_species_name);
                if (spData == null) {
                    throw new Exception("No species in ensembl registry called: " + ensembl_genus_species_name);
                }
            }
            spData.getAliases().add(alias);
            this.ensemblNameForAliasHash.put(alias, ensembl_genus_species_name);
            return true;
        } catch (Exception e) {
            throw new Exception("Failed to add alias: " + alias + " for species: " + ensembl_genus_species_name);
        }

    }

    @Override
    public Properties getConfigProperties() {
        return configProperties;
    }

    @Override
    public StringBuffer printRegistryReport() {
        StringBuffer out = new StringBuffer();

        if (this.regConfig==null) {
            out.append("This Registry object has not been configured"+newline);
            return out;
        }

        out.append("Registry for: " + this.dbConnection.getConfigurationProperties().getProperty("url")+newline);

        out.append(dbConnection.report()+newline);

        out.append(" Current Ensembl Release: " + this.mostRecentEnsemblVersion+newline);
        out.append(" Current EnsemblGenomes Release: " + this.mostRecentGenomesVersion+newline);
        out.append(" Current DB Version: " + this.getMostRecentDBVersion()+newline);

        if (!this.alreadyInitialized) {
            out.append("This configured Registry has not yet been initialized"+newline);
            return out;
        }

        if (tooNewDB.toString() != null && !tooNewDB.toString().isEmpty()) {
            out.append("Some database releases were too recent for the current JEnsembl configuration:"
                    + newline + tooNewDB.toString());
        }
        if (unknownDB.toString() != null && !unknownDB.toString().isEmpty()) {
            out.append("Some database releases were not known in the current JEnsembl configuration:"
                    + newline + unknownDB.toString());
        }


        out.append(newline+"___________________________"+newline);
        out.append("Species in Ensembl (with aliases):"+newline);
        out.append("\tSingle Species"+newline+"______________"+newline);

        for (DBSpecies s : this.speciesHash.values()) {
            out.append(s.getSpeciesBinomial() + " [" + s.getDatabaseStyleName() + "] (");
            for (String st : s.getAliases()) {
                out.append(st + ", ");
            }
            try {
                out.append(")"+newline+"\tCurrent Core Database: " + s.getCurrentCoreDatabase().getdBName()  +newline);

                for (FeatureType key : ((DBSingleSpeciesCoreDatabase) s.getCurrentCoreDatabase()).getBuildLevels().keySet()) {
                    out.append(key.toString() + " : " + ((DBSingleSpeciesCoreDatabase) s.getCurrentCoreDatabase()).getBuildLevels().get(key) + "\n");
                }

            } catch (Exception e) {
                out.append(")"+newline+"\tNO Current Core Database"+newline);
            }
        }


        out.append(newline+"_____________"+newline+"CollectionDatabase Species"+newline+"______________"+newline);

        if (this.collectionspeciesHash.values().isEmpty()) {
            out.append(newline+"NONE"+newline);
        }

        for (DBCollectionSpecies s : this.collectionspeciesHash.values()) {
            out.append(s.getSpeciesBinomial() + " [" + s.getDatabaseStyleName() + "] (");
            for (String st : s.getAliases()) {
                out.append(st + ", ");
            }
            try {
                out.append(")"+newline+"\tCurrent Core Database: " + s.getCurrentCoreDatabase().getdBName() +newline);
            } catch (Exception e) {
                out.append(")"+newline+"\tNO Current Core Database"+newline);
            }
        }



        out.append("\n____________________\n\n\nMulti Species Databases: ");
        out.append("\n________________________\n\n");
        for (EnsemblDBType t : EnsemblDBType.getSpeciesComparisonDatabaseTypes()) {
            out.append(t.toString() + ":\n");
            for (Database d : this.getDatabasesByType(t)) {
                out.append("\t" + d.getdBName() + "\n");
            }
        }


        out.append("\nSingleSpecies Database details: ");
        out.append("\n________________________\n\n");

        for (DBSpecies sp : this.speciesHash.values()) {
            String id = sp.getSpeciesBinomial();
            out.append(id + "\n__________________\n");


            out.append("\n\tCurrent Versions:\n");
            for (Database d : sp.getDatabasesByVersion(this.getMostRecentDBVersion())) {
                out.append("\t\t" + d.getdBName() + "\n");
            }
            out.append("\tBy Type:\n");
            for (EnsemblDBType t : EnsemblDBType.getSingleSpeciesDatabaseTypes()) {

                out.append("\t" + t.toString() + "\n");
                if (sp.getDatabasesByType(t).isEmpty()) {
                    out.append("\t\tNONE\n");
                } else {
                    for (Database d : sp.getDatabasesByType(t)) {
                        out.append("\t\t" + d.getdBName() + "\n");
                    }
                }
            }

        }
        out.append("\nCollectionSpecies Database details: ");
        out.append("\n________________________\n\n");

        if (this.collectionspeciesHash.values().isEmpty()) {
            out.append("\nNONE\n");
        }

        for (DBCollectionSpecies sp : this.collectionspeciesHash.values()) {
            String id = sp.getSpeciesBinomial();
            out.append(id + "\n__________________\n");


            out.append("\n\tCurrent Versions:\n");
            for (Database d : sp.getDatabasesByVersion(this.getMostRecentDBVersion())) {
                out.append("\t\t" + d.getdBName() + "\n");
            }
            out.append("\tBy Type:\n");
            for (EnsemblDBType t : EnsemblDBType.getCollectionDatabaseTypes()) {

                out.append("\t" + t.toString() + "\n");
                if (sp.getDatabasesByType(t).isEmpty()) {
                    out.append("\t\tNONE\n");
                } else {
                    for (Database d : sp.getDatabasesByType(t)) {
                        out.append("\t\t" + d.getdBName() + "\n");
                    }
                }
            }
        }

        out.append("\nCore Collection Databases: Species and build information");
        out.append("\n________________________________________________________\n");

        if (this.collectionDatabases.isEmpty()) {
            out.append("\nNONE\n");
        }

        for (DBCollectionDatabase d : this.collectionDatabases) {
            if (d.getType().equals(EnsemblDBType.collection_core)) {

                out.append("\nCORECOLLECTION DB: " + d.dBName + "\n");

                for (DBCollectionSpecies species : d.getCollection().getSpecies()) {

                    DBCollectionCoreDatabase db = (DBCollectionCoreDatabase) d;
                    out.append("\tSPECIES: " + species.commonName + "\n");


                    for (FeatureType key : db.getBuildLevels(species).keySet()) {
                        out.append("\t\t" + key.toString() + " : " + db.getBuildLevels(species).get(key) + "\n");
                    }


                }
            }
        }


        return out;
    }

    @Override
    public DBSpecies getSpeciesByAlias(String alias_or_name) {
        if (alias_or_name==null||alias_or_name.isEmpty()) {
            return null;
        }

        DBSpecies out = null;
        out = (this.getEnsemblNameForAlias(alias_or_name) != null
                && this.speciesHash.get(this.getEnsemblNameForAlias(alias_or_name)) != null)
                ? this.speciesHash.get(this.getEnsemblNameForAlias(alias_or_name))
                : null;
        if (out == null) {
            out = (this.getEnsemblNameForAlias(alias_or_name) != null
                    && this.collectionspeciesHash.get(this.getEnsemblNameForAlias(alias_or_name)) != null)
                    ? this.collectionspeciesHash.get(this.getEnsemblNameForAlias(alias_or_name))
                    : null;
        }

        return out;
    }

    @Override
    public DBSpecies getSpeciesByEnsemblName(String ensemblName) {
        if (ensemblName==null||ensemblName.isEmpty()) {
            return null;
        }

        DBSpecies out = null;
        out = this.speciesHash.get(ensemblName);
        if (out == null) {
            out = this.collectionspeciesHash.get(ensemblName);
        }

        return out;
    }

    @Override
    public String getMostRecentEnsemblVersion() {
        return mostRecentEnsemblVersion;
    }

    @Override
    public DBCollection getCollection(String name, String dbVersion) {
        DBCollection out = null;
        if (name==null||name.isEmpty()||dbVersion==null||dbVersion.isEmpty()) {
            return out;
        }
        if (this.collectionNameDBVersionHash.containsKey(name)) {
            if (this.collectionNameDBVersionHash.get(name).containsKey(dbVersion)) {
                out = this.collectionNameDBVersionHash.get(name).get(dbVersion);
            }
        }
        return out;
    }

    private void registerDatabase(DBDatabase db) {
        if (db instanceof DBCollectionDatabase) {
            this.collectionDatabases.add((DBCollectionDatabase) db);
            if (db.intSchemaVersion > this.highestEnsemblVersion) {
                this.highestEnsemblVersion = db.intSchemaVersion;
            }
            if (db.intDBVersion > this.highestVersion) {
                this.highestVersion = db.intDBVersion;
            }
        } else if (db instanceof DBSingleSpeciesDatabase) {
            this.singleSpeciesDatabases.add((DBSingleSpeciesDatabase) db);
            if (db.intSchemaVersion > this.highestEnsemblVersion) {
                this.highestEnsemblVersion = db.intSchemaVersion;
            }
            if (db.intDBVersion > this.highestVersion) {
                this.highestVersion = db.intDBVersion;
            }
        } else if (db instanceof DBComparisonDatabase) {
            this.comparisonDatabases.add((DBComparisonDatabase) db);

            if (db.getType().equals(EnsemblDBType.compara)) {

            if (!this.comparaDBByDivision.containsKey(((DBComparisonDatabase)db).getComparisonDivision())) {
                this.comparaDBByDivision.put(((DBComparisonDatabase)db).getComparisonDivision(),
                        new HashMap<String,DBComparisonDatabase> ());
            }
            this.comparaDBByDivision.get(((DBComparisonDatabase)db).getComparisonDivision()).put(
                    db.getDBVersion(),(DBComparisonDatabase) db);

            }

        }
        this.allDatabases.add(db);
    }

    @Override
    public Collection<DBCollection> getCollectionRegistriesByName(String name) {
        if (name==null||name.isEmpty()) {
            return new ArrayList<DBCollection>();
        }
        if (collectionNameDBVersionHash.containsKey(name)) {
            return new ArrayList<DBCollection>(collectionNameDBVersionHash.get(name).values());
        } else {
            return new ArrayList<DBCollection>();
        }
    }

    @Override
    public Collection<DBCollection> getCollectionsByDBVersion(String dbVersion) {

        Collection<DBCollection> out = new ArrayList<DBCollection>();
        if (dbVersion==null||dbVersion.isEmpty()) {
            return out;
        }

        for (String name : this.collectionNameDBVersionHash.keySet()) {

            if (this.collectionNameDBVersionHash.get(name).containsKey(dbVersion)) {
                out.add(this.collectionNameDBVersionHash.get(name).get(dbVersion));
            }
        }
        return out;
    }

    @Override
    public List<DBCollection> getCollections() {
        return collections;
    }

    @Override
    public Collection<DBSpecies> getSpecies() {
        return new ArrayList<DBSpecies>(speciesHash.values());
    }

    @Override
    public Collection<DBCollectionSpecies> getCollectionSpecies() {
        return new ArrayList<DBCollectionSpecies>(collectionspeciesHash.values());
    }

    @Override
    public void addSpecies(String dbStyleName, Species species) {

        if (dbStyleName==null || dbStyleName.isEmpty() || species==null) {
            return;
        }

        if (species instanceof DBCollectionSpecies) {
            this.collectionspeciesHash.put(dbStyleName, (DBCollectionSpecies) species);
        } else {
            this.speciesHash.put(dbStyleName, (DBSpecies) species);
        }

    }

    @Override
    public String getMostRecentDBVersion() {

        if (this.dbConnection != null 
                && this.dbConnection.getDatasourceStyle().equalsIgnoreCase("ensemblgenomes") ) {
            return this.mostRecentGenomesVersion;
        } else {
            return this.mostRecentEnsemblVersion;
        }
    }

    public String getBriefRegistryReport() {
        StringBuffer out = new StringBuffer();
        if (this.regConfig==null) {
            out.append("This Registry object has not been configured"+newline);
            return out.toString();
        }
        out.append("Registry for: " + this.dbConnection.getConfigurationProperties().getProperty("url") + newline);
        out.append("Highest configured release : " + this.mostRecentEnsemblVersion + newline);
        out.append("Known schema versions: " + newline);
        for (Integer i : knownSchemaVersions) {
            out.append("\t" + i + newline);
        }
        out.append(dbConnection.report());

        if (!this.alreadyInitialized) {
            out.append("This configured Registry has not yet been initialized"+newline);
            return out.toString();
        }

        if (tooNewDB.toString() != null && !tooNewDB.toString().isEmpty()) {
            out.append("Some database releases were too recent for the current JEnsembl configuration:"
                    + newline + tooNewDB.toString());
        }
        if (unknownDB.toString() != null && !unknownDB.toString().isEmpty()) {
            out.append("Some database releases were not known in the current JEnsembl configuration:"
                    + newline + unknownDB.toString());
        }



        return out.toString();
    }

    public TreeSet<Integer> getKnownSchemaVersions() {
        return knownSchemaVersions;
    }

    @Override
    public int getHighestEnsemblSchemaVersion() {
        return highestEnsemblVersion;
    }

    @Override
    public int getHighestDBVersion() {
        return highestVersion;
    }

    @Override
    public DataSource getDatasourceType() {
        return datasourceType;
    }

    @Override
    public DBDatabase makeDatabase(String db_name) throws ConfigurationException {

        if (db_name==null||db_name.isEmpty()) {
            throw new ConfigurationException("Invalid database name: "+db_name);
        }
 
        EnsemblDBType t = null;

        for (EnsemblDBType et : EnsemblDBType.getAllDatabaseTypes()) {

            if (db_name.contains("_collection_") && db_name.matches(new String(".+_" + et.toString() + "_\\d+.*"))) {
                t=et;
                break;
            }

            else if (db_name.matches(new String(".+_" + et.toString() + "_\\d+.*"))
                    || (et.toString().startsWith("ensembl_")&& db_name.contains(et.toString() + "_"))) {
                t = et;
            }
        }

        if (t==null) {
            throw new ConfigurationException("Unrecognized database type for database name: "+db_name);
        }


        try {
            if (EnsemblDBType.getSingleSpeciesDatabaseTypes().contains(t)) {
                if (t.equals(EnsemblDBType.core)) {
                    return new DBSingleSpeciesCoreDatabase(db_name, t, this);
                } else {
                    //haven't implemented othersubtypes yet
                    return new DBSingleSpeciesDatabase(db_name, t, this);
                }
            } else if (EnsemblDBType.getCollectionDatabaseTypes().contains(t)) {
                if (t.equals(EnsemblDBType.collection_core)) {
                  return new DBCollectionCoreDatabase(db_name, t, this);
                } else {
                  //haven't implemented other subtypes yet
                   return new DBCollectionDatabase(db_name, t, this);
                }
            } else if (EnsemblDBType.getSpeciesComparisonDatabaseTypes().contains(t)) {
                return new DBComparisonDatabase(db_name, t, this);
            } else {
                throw new Exception("Unrecognized database type for database name: " + db_name);
            }
        } catch (Exception e) {
            throw new ConfigurationException(e.getMessage());
        }

    }

    //DAO methods

    private List<String> getDatabases() throws DAOException {

        List<String> outList = null;
        SqlSession session = null;

        try {
            session = sqlMapper.openSession();
            DatabaseMapper dm = session.getMapper(DatabaseMapper.class);
            outList = dm.getAllDatabaseNames();
        } catch (Exception e) {
            throw new DAOException("Failed to interrogate all Database Names", e);
        } finally {
            if (session != null) {
                session.close();
            }
        }

        if (outList == null) {
            /* A null tempList is a perfectly valid outcome to return */
            return new ArrayList<String>();
        }

        return outList;
    }

    private DBCollection getCollectionSpeciesProperties(CollectionCoreDatabase database) throws DAOException {

        DBCollection collection = null;
        DBCollectionCoreDatabase myDatabase = (DBCollectionCoreDatabase) database;
        String dBName = myDatabase.dBName;
        String schemaVersion = myDatabase.getSchemaVersion();
        String dbVersion = myDatabase.getDBVersion();
        List<DBCollectionSpecies> out = new ArrayList<DBCollectionSpecies>();
        TreeMap<Integer, DBCollectionSpecies> localSpp = new TreeMap<Integer, DBCollectionSpecies>();
        collection = new DBCollection(myDatabase);

        List<HashMap> results = null;
        SqlSession session = null;

        try {
            session = sqlMapper.openSession();
            DatabaseMapper dm = session.getMapper(DatabaseMapper.class);
            results = dm.getSpeciesFromCollection(dBName);
        } catch (Exception e) {
            throw new DAOException("Failed to interrogate species properties for " + dBName, e);
        } finally {
            if (session != null) {
                session.close();
            }
        }

        if ((results == null) || (results.isEmpty() == true)) {
            /*  return empty collectiont*/
            return collection;
        }

        HashMap<Integer, HashMap<String, String>> buildLevelsHash = new HashMap<Integer, HashMap<String, String>>();

        for (Object o : results) {

            Integer id = null;
            DBCollectionSpecies species = null;

            HashMap m = (HashMap) o;
            m.put("schemaVersion", schemaVersion);

            id = (Integer) m.get("species_id");

            // if we've already made species with this id  for this collection
            //use it
            if (localSpp.containsKey(id)) {
                species = localSpp.get(id);
            }
            //we haven't got a species with this id in the collection already
            //- so make one, add it, and use it
            else {
                try {
                    species = new DBCollectionSpecies(myDatabase);
                } catch (ConfigurationException ex) {
                    //what do we do here??
                }
                if (species==null) {
                    continue;
                }
                species.setIDForVersion(id, dbVersion);
                localSpp.put(id, species);
            }

            if (!((String) m.get("meta_key")).endsWith("build.level")) {
                species.setProperty(m);
            } else {

                if (buildLevelsHash.containsKey(id)) {
                    buildLevelsHash.get(id).put((String) m.get("meta_key"), (String) m.get("meta_value"));
                } else {
                    buildLevelsHash.put(id, new HashMap<String, String>());
                    buildLevelsHash.get(id).put((String) m.get("meta_key"), (String) m.get("meta_value"));
                }
            }

            if (species.getCurrentCoreDatabase() == null
                    && myDatabase.getType().equals(EnsemblDBType.collection_core)
                    && dbVersion.equalsIgnoreCase(this.getMostRecentDBVersion())) {
                species.setCurrentCoreDatabase(myDatabase);
            }
        }

        for (DBCollectionSpecies s : localSpp.values()) {
            //this cant be used cos we havent populated the alias look up yet!
            //if (this.registry.getSpeciesByAlias(s.getDatabaseStyleName()) == null) {


            //ensembl have changed the case of the first letter of the dbname between v58 & v59
            //therefore the code below fails to rationalize species between collections 58 and 59
            //i am just going to live with this for the time being bcause the dbname is also used as the comparaname
            //and is also different between 58 and 59
            //hopefully from 59 things will be stable



            //give this to the reg

            if (this.getSpeciesByEnsemblName(s.getDatabaseStyleName()) == null) {
                //we've made a new species so add itto the registry
                this.addSpecies(s.getDatabaseStyleName(), s);
                //and add it to this collection
                collection.addSpecies(s);
            } else {
                //we've already got this species in the registry - so modify it there
                DBCollectionSpecies existingSpecies = (DBCollectionSpecies) this.getSpeciesByEnsemblName(s.getDatabaseStyleName());
                existingSpecies.getAliases().addAll(s.getAliases());
                existingSpecies.setIDForVersion(s.getDBSpeciesID(dbVersion), dbVersion);
                existingSpecies.addDatabase(myDatabase);
                if (existingSpecies.getCurrentCoreDatabase() == null
                        && s.getCurrentCoreDatabase() != null) {
                    existingSpecies.setCurrentCoreDatabase(s.getCurrentCoreDatabase());
                }
                //and add it to this collection
                collection.addSpecies(existingSpecies);
            }
        }

        myDatabase.setCollection(collection);

        for (DBSpecies s : collection.getSpecies()) {
            DBCollectionSpecies sp = (DBCollectionSpecies) s;
            if (buildLevelsHash.containsKey(sp.getDBSpeciesID(dbVersion))) {
                myDatabase.setBuildLevels(sp, buildLevelsHash.get(sp.getDBSpeciesID(dbVersion)));
                HashMap<String, String> map = buildLevelsHash.get(sp.getDBSpeciesID(dbVersion));
            }
        }

        return collection;
    }

    private List<String> getAllCoreDatabasesForVersion(int dbVersion) throws DAOException {

        int release = dbVersion;
        List<String> outList = null;
        SqlSession session = null;

        try {
            session = sqlMapper.openSession();
            DatabaseMapper dm = session.getMapper(DatabaseMapper.class);
            outList = dm.getAllCoreDatabaseNames(release);
        } catch (Exception e) {
            throw new DAOException("Failed to interrogate all core Database Names for release " + release, e);
        } finally {
            if (session != null) {
                session.close();
            }
        }

        if (outList == null || outList.isEmpty()) {
            /* A null List is a perfectly valid outcome to return */
            return null;
        }

        return outList;
    }

    public void setCoreDBBuildLevels(SingleSpeciesCoreDatabase database) throws DAOException {
        
        List<HashMap> tempList = null;

        SqlSession session = null;

        try {
            session = sqlMapper.openSession();
            DatabaseMapper dm = session.getMapper(DatabaseMapper.class);
            tempList = dm.getSpeciesProperties(database.getdBName());
        } catch (Exception e) {
            throw new DAOException("Failed to interrogate db "+database.getdBName()+" for build levels", e);
        } finally {
            if (session != null) {
                session.close();
            }
        }
        
        if ((tempList == null) || (tempList.isEmpty() == true)) {
            return ;
        }
        
        Iterator it = tempList.iterator();

        while (it.hasNext()) {
            HashMap row = (HashMap) it.next();

            String key = (String) row.get("key");
            String value = (String) row.get("value");

            if (key==null || key.isEmpty() || value==null) {
                continue;
            }

            if (key.endsWith("build.level")) {
                database.setBuildLevel(key, value);
            } 

        }

    }

    private DBSpecies getSpeciesProperties(SingleSpeciesCoreDatabase database) throws DAOException {

        DBSpecies spData = new DBSpecies((DBSingleSpeciesCoreDatabase) database);

        List<HashMap> tempList = null;

        SqlSession session = null;

        try {
            session = sqlMapper.openSession();
            DatabaseMapper dm = session.getMapper(DatabaseMapper.class);
            tempList = dm.getSpeciesProperties(database.getdBName());
        } catch (Exception e) {
            throw new DAOException("Failed to interrogate species properties for " + spData.getDatabaseStyleName(), e);
        } finally {
            if (session != null) {
                session.close();
            }
        }

        //for pre dbVersion 59 releases, the meta table is not guaranteed
        //to hold the species.production_name, species.short_name and
        //species.scientific_name meta_keys

        //pre schemaVersion 59 the compara name was always = Species Binomial
        //post schemaVersion 58 the compara name - production name

        if (spData.getDatabaseStyleName().isEmpty()) {
            spData.setDatabaseStyleSpeciesName(database.getDbSpeciesName());
            spData.getAliases().add(spData.getDatabaseStyleName());
        }

        if (spData.getSpeciesBinomial().isEmpty()) {
            String s = database.getDbSpeciesName();
            spData.setSpeciesBinomial( s.replaceFirst( s.substring(0, 1), (s.substring(0,1)).toUpperCase() ).replace("_"," "));
            spData.getAliases().add(spData.getSpeciesBinomial());
        }

        if ((tempList == null) || (tempList.isEmpty() == true)) {
            return spData;
        }

        Iterator it = tempList.iterator();

        while (it.hasNext()) {
            HashMap row = (HashMap) it.next();

            String key = (String) row.get("key");
            String value = (String) row.get("value");

            if (key==null || key.isEmpty() || value==null) {
                continue;
            }

            if (key.equals("species.stable_id_prefix")) {
                spData.setEnsemblStablePrefix(value);
            } else if (key.equals("species.production_name")) {
                spData.setDatabaseStyleSpeciesName(value);
                spData.getAliases().add(value);
            } else if (key.equals("species.common_name")) {
                spData.getAliases().add(value);
                spData.setCommonName(value);
            } else if (key.equals("species.taxonomy_id")) {
                spData.getAliases().add(value);
                spData.setTaxonomyID(value);
            } else if (key.equals("species.short_name")) {
                spData.getAliases().add(value);
                spData.setShortName(value);
            } else if (key.equals("species.alias")
                    || key.equals("species.ensembl_common_name")
                    || key.equals("species.ensembl_alias_name")) {
                spData.getAliases().add(value);
            } else if (key.endsWith("build.level")) {
                database.setBuildLevel(key, value);
            } else if (key.equals("species.division")) {
                spData.setComparaDivision(value);
            }

        }

        if (Integer.parseInt(database.getSchemaVersion()) >= 59   ) {
            spData.setComparaName(spData.getDatabaseStyleName());
        } else {
            spData.setComparaName(spData.getSpeciesBinomial());
        }

        return spData;
    }

    @Override
    public String getAssembly(SingleSpeciesDatabase database) throws DAOException {

        if (database== null || database.getdBName()==null || database.getdBName().isEmpty()) {
            return "";
        }

        //for current Databases this query will be cached by mybatis
        //cos it was used to make the species

        List<HashMap> tempList = null;

        SqlSession session = null;

        try {
            session = sqlMapper.openSession();
            DatabaseMapper dm = session.getMapper(DatabaseMapper.class);
            tempList = dm.getSpeciesProperties(database.getdBName());
        } catch (Exception e) {
            throw new DAOException("Failed to retrieve assembly name for " + database.getdBName(), e);
        } finally {
            if (session != null) {
                session.close();
            }
        }



        if ((tempList == null) || (tempList.isEmpty() == true)) {
            return "";
        }

        Iterator it = tempList.iterator();

        while (it.hasNext()) {
            HashMap row = (HashMap) it.next();

            String key = (String) row.get("key");
            String value = (String) row.get("value");

            if (key==null || key.isEmpty() || value==null) {
                continue;
            }

            if (key.equals("assembly.name")) {
                return value;
            }

        }

        return "";
    }

    @Override
    public HashMap<CollectionSpecies,String> getAssemblies(CollectionDatabase database) throws DAOException {

        HashMap<CollectionSpecies,String> out = new HashMap<CollectionSpecies,String>();

        if (database== null || database.getdBName()==null || database.getdBName().isEmpty()
                || database.getCollection()==null || database.getCollection().getSpecies().isEmpty()) {
            return out;
        }

        TreeSet<? extends CollectionSpecies> spp = database.getCollection().getSpecies();

        HashMap<Integer, DBCollectionSpecies> localSpp = new HashMap<Integer, DBCollectionSpecies>();

        for (CollectionSpecies sp : spp) {
            localSpp.put(sp.getDBSpeciesID(database.getDBVersion()), (DBCollectionSpecies) sp);
        }

        String dBName = database.getdBName();
        List<HashMap> results = null;
        SqlSession session = null;

        try {
            session = sqlMapper.openSession();
            DatabaseMapper dm = session.getMapper(DatabaseMapper.class);
            results = dm.getSpeciesFromCollection(dBName);
        } catch (Exception e) {
            throw new DAOException("Failed to get assembly names for " + dBName, e);
        } finally {
            if (session != null) {
                session.close();
            }
        }

        if ((results == null) || (results.isEmpty() == true)) {
            /*  return empty collectiont*/
            return out;
        }


        for (Object o : results) {

            Integer id = null;
            DBCollectionSpecies species = null;

            HashMap m = (HashMap) o;

            id = (Integer) m.get("species_id");

            // if we've already made species with this id  for this collection
            //use it
            if (localSpp.containsKey(id)) {
                species = localSpp.get(id);
            } else {
                continue;
            }

            if ( m.get("meta_key").equals("assembly.name") ) {
                out.put(species, (String) m.get("meta_value"));
            }
        }


        return out;
    }


}
