/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */

package uk.ac.roslin.ensembl.dao.database;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeMap;
import java.util.logging.Level;
import java.util.logging.Logger;
import uk.ac.roslin.ensembl.datasourceaware.core.DAChromosome;
import uk.ac.roslin.ensembl.dao.database.factory.DBDAOSingleSpeciesCoreFactory;
import uk.ac.roslin.ensembl.dao.database.factory.DBDAOSingleSpeciesFactory;
import uk.ac.roslin.ensembl.exception.ConfigurationException;
import uk.ac.roslin.ensembl.exception.DAOException;
import uk.ac.roslin.ensembl.config.EnsemblDBType;
import uk.ac.roslin.ensembl.datasourceaware.core.DACoordinateSystem;
import uk.ac.roslin.ensembl.model.ObjectType;
import uk.ac.roslin.ensembl.model.core.CoordinateSystem;
import uk.ac.roslin.ensembl.model.database.SingleSpeciesCoreDatabase;
import uk.ac.roslin.ensembl.config.EnsemblCoordSystemType;
import uk.ac.roslin.ensembl.config.FeatureType;
import uk.ac.roslin.ensembl.model.database.Registry;

/**
 *
 * @author paterson
 */
public class DBSingleSpeciesCoreDatabase extends DBSingleSpeciesDatabase implements SingleSpeciesCoreDatabase {

    private List<DACoordinateSystem> coordSystems = new ArrayList<DACoordinateSystem>();
    private TreeMap<Integer, DACoordinateSystem> rankedCoordSystemsHash = new TreeMap<Integer, DACoordinateSystem>();
    private TreeMap<Integer, DACoordinateSystem> idCoordSystemsHash = new TreeMap<Integer, DACoordinateSystem>();
    private Integer chromosomeLevelID = null;
    private Integer sequenceLevelID = null;
    private Integer topLevelID = null;
    private DACoordinateSystem chromosomeLevelCS = null;
    private DACoordinateSystem sequenceLevelCS = null;
    private DACoordinateSystem topLevelCS = null;
    private HashMap<FeatureType, String> defaultBuildLevels = new HashMap<FeatureType, String>();
    private HashMap<FeatureType, HashMap<DACoordinateSystem, Integer>> featureCSHash = null;
    private HashMap<DACoordinateSystem, List<FeatureType>> CSFeatureHash = null;

    public DBSingleSpeciesCoreDatabase(String db_name, EnsemblDBType type, Registry registry) throws ConfigurationException {
        super(db_name, type, registry);
    }

    @Override
    public DBDAOSingleSpeciesCoreFactory getCoreFactory() {
        if (this.factory == null) {
            try {
                this.factory = DBDAOSingleSpeciesFactory.makeFactory(this);
            } catch (Exception ex) {
            }
        }

        if (this.factory != null) {
            return (DBDAOSingleSpeciesCoreFactory) this.factory;
        } else {
            return null;
        }
    }

    
    private List<DACoordinateSystem> fetchCS() throws DAOException {
        List<DACoordinateSystem> coords = new ArrayList<DACoordinateSystem>();
        //can throw DAOException
        if ( this.getCoreFactory()!=null) {
            coords = this.getCoreFactory().getCoordinateSystemDAO().
                getCoordinateSystems();
        } else {
           return coords;
        }        
        return coords;
    }
    
    private void setCoordinateSytems(List<DACoordinateSystem> coords) throws DAOException {

        this.coordSystems.clear();
        this.coordSystems.addAll(coords);

        //set  the ids for the top, chromosome and sequence levels

        for (DACoordinateSystem cs : coordSystems) {

            this.rankedCoordSystemsHash.put(cs.getRank(), cs);
            this.idCoordSystemsHash.put(cs.getId(), cs);

            if (cs.getRank().equals(1)) {
                this.topLevelCS = cs;
                this.topLevelID = cs.getId();
                if (cs.getType() == EnsemblCoordSystemType.chromosome) {
                    this.chromosomeLevelCS = cs;
                    this.chromosomeLevelID = cs.getId();
                }
            }

            if (cs.isSequenceLevel()) {
                this.sequenceLevelCS = cs;
                this.sequenceLevelID = cs.getId();
            }
        }

        //if the top rank wasnt chromosome level, find the highest ranked chromosome level

        if (this.chromosomeLevelID == null) {


            Iterator<Integer> it = this.rankedCoordSystemsHash.keySet().iterator();
            while (it.hasNext()) {
                DACoordinateSystem cs = this.rankedCoordSystemsHash.get(it.next());
                if (cs.getType() == EnsemblCoordSystemType.chromosome) {
                    this.chromosomeLevelID = cs.getId();
                    break;
                }
            }
        }


    }

    private void lazyLoadCoordinateSytems() throws DAOException {

        this.coordSystems.clear();

        List<DACoordinateSystem> coords = this.fetchCS();

        if (!coords.isEmpty()) {
               this.setCoordinateSytems(coords);
        }

        //set nil values to prevent repeating lazy load

        if (this.chromosomeLevelID == null) {
            this.chromosomeLevelID = 0;
        }
        if (this.topLevelID == null) {
            this.topLevelID = 0;
        }
        if (this.sequenceLevelID == null) {
            this.sequenceLevelID = 0;
        }
       
    }

    @Override
    public DACoordinateSystem getCSByID(Integer id) throws DAOException {
        DACoordinateSystem out = null;
        if (topLevelID == null && this.coordSystems.isEmpty()) {

                //lazyload can throw DAOException
                lazyLoadCoordinateSytems();

        }

        out = this.idCoordSystemsHash.get(id);

        return out;
    }

    @Override
    public DACoordinateSystem getTopLevelCoordSystem() throws DAOException {
        if (topLevelID == null && this.coordSystems.isEmpty()) {
            //lazyload can throw DAOException
            lazyLoadCoordinateSytems();
        }
        return topLevelCS;
    }

    @Override
    public DACoordinateSystem getChromosomeLevelCoordSystem() throws DAOException {
        if (chromosomeLevelID == null && this.coordSystems.isEmpty()) {
            //lazyload can throw DAOException
            lazyLoadCoordinateSytems();
        }
        return chromosomeLevelCS;
    }

    @Override
    public DACoordinateSystem getSequenceLevelCoordSystem() throws DAOException {
        if (sequenceLevelID == null && this.coordSystems.isEmpty()) {
            //lazyload can throw DAOException
            lazyLoadCoordinateSytems();
        }
        return sequenceLevelCS;
    }

    @Override
    public DAChromosome getChromosomeByName(String name) throws DAOException {
        if (this.getCoreFactory()!=null){
            return this.getCoreFactory().getChromosomeDAO().getChromosomeByName(name);
        } else {
            return null;
        }

    }

    @Override
    public List<DAChromosome> getChromosomes() throws DAOException {
        if (this.getCoreFactory()!=null){
            return this.getCoreFactory().getChromosomeDAO().getChromosomes();
         } else {
            return new ArrayList<DAChromosome>();
        }
    }

    private void setFeatureCSHash() throws DAOException {

        if (sequenceLevelID == null && this.coordSystems.isEmpty()) {
            //lazyload can throw DAOException
            lazyLoadCoordinateSytems();
        }

        featureCSHash = new HashMap<FeatureType, HashMap<DACoordinateSystem, Integer>>();
        CSFeatureHash = new HashMap<DACoordinateSystem, List<FeatureType>>();

        for (DACoordinateSystem c : this.coordSystems) {
            CSFeatureHash.put(c, new ArrayList<FeatureType>());

        }


        if (this.getCoreFactory() != null) {

            try {

                this.getCoreFactory().getCoordinateSystemDAO().setFeatureCS();

            } catch (DAOException e) {
                featureCSHash = null;
                CSFeatureHash = null;
                throw e;
            }
        }

    }

    @Override
    public Set<DACoordinateSystem> getCSForFeature(ObjectType featureType) throws DAOException {

            if (this.featureCSHash == null) {
                this.setFeatureCSHash();
            }


        if (this.featureCSHash != null
                && this.featureCSHash.containsKey(featureType)) {
            return this.featureCSHash.get(featureType).keySet();
        } else {

            return null;
        }
    }

    @Override
    public List<FeatureType> getFeaturesForCS(CoordinateSystem coordSys) throws DAOException{


            if (this.featureCSHash == null) {
                this.setFeatureCSHash();
            }


        if (this.CSFeatureHash != null
                && this.CSFeatureHash.containsKey(coordSys)) {
            return this.CSFeatureHash.get(coordSys);
        } else {
            return null;
        }
    }

    @Override
    public Integer getMaxLengthForFeature(ObjectType featureType, CoordinateSystem cs) throws DAOException {

        if ( featureType == null || cs == null ) {
            return null;
        }

            if (this.featureCSHash == null) {
                this.setFeatureCSHash();
            }



        if (this.featureCSHash != null
                && this.featureCSHash.containsKey(featureType)
                && this.featureCSHash.get(featureType).containsKey(cs)) {
            return this.featureCSHash.get(featureType).get(cs);
        } else {
            return null;
        }

    }

    @Override
    public void addFeatureCS(String featureType, Integer csID, Integer maxLength) {

        FeatureType type = null;
        DACoordinateSystem cs = null;

        type = FeatureType.getFeatureType(featureType);
        cs = this.idCoordSystemsHash.get(csID);

        if (type == null || cs == null) {
            return;
        } else {
            if (!this.featureCSHash.containsKey(type)) {
                this.featureCSHash.put(type, new HashMap<DACoordinateSystem, Integer>());
            }
            this.featureCSHash.get(type).put(cs, maxLength);

            this.CSFeatureHash.get(cs).add(type);

        }


    }

    @Override
    public void setBuildLevel(String featureKey, String level) {
        String temp = featureKey.replace("build.level", "");

        if (FeatureType.getFeatureType(temp) != null) {
            this.defaultBuildLevels.put(FeatureType.getFeatureType(temp), level);
        }
    }

    @Override
    public String getBuildLevel(String featureType) throws DAOException{
        if (FeatureType.getFeatureType(featureType) != null) {
            if (defaultBuildLevels.isEmpty()) {
                this.getRegistry().setCoreDBBuildLevels(this);
            }
            return this.defaultBuildLevels.get(FeatureType.getFeatureType(featureType));
        } else {
            return null;
        }
    }

    @Override
    public DACoordinateSystem getBuildCoordSystem(String featureType) throws DAOException {

            if (FeatureType.getFeatureType(featureType) != null) {
                if (defaultBuildLevels.isEmpty()) {
                    this.getRegistry().setCoreDBBuildLevels(this);
                }
                String level = this.defaultBuildLevels.get(FeatureType.getFeatureType(featureType));
                if (level != null && level.equalsIgnoreCase("toplevel")) {
                    return this.getTopLevelCoordSystem();
                } else {
                    return null;
                }

//                //are there other values possible?
//                else if (level.equalsIgnoreCase("chromosomelevel")) {
//                    return this.getChromosomeLevelCoordSystem();
//                } else if (level.equalsIgnoreCase("sequencelevel")) {
//                    return this.getSequenceLevelCoordSystem();
//                }
            }

        return null;

    }

    //dont really want this public - just for testing
    public HashMap<FeatureType, String> getBuildLevels() {
        if (defaultBuildLevels.isEmpty()) {
            try {
                lazyLoadCoordinateSytems();
            } catch (DAOException ex) {
                
            }
        }
        return defaultBuildLevels;
    }
}
