/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */
package uk.ac.roslin.ensembl.dao.database;

import java.util.HashMap;
import java.util.List;
import java.util.TreeMap;
import java.util.TreeSet;
import uk.ac.roslin.ensembl.datasourceaware.core.DAChromosome;
import uk.ac.roslin.ensembl.datasourceaware.core.DAGene;
import uk.ac.roslin.ensembl.exception.DAOException;
import uk.ac.roslin.ensembl.model.core.Species;
import uk.ac.roslin.ensembl.model.database.CoreDatabase;
import uk.ac.roslin.ensembl.model.database.Database;
import uk.ac.roslin.ensembl.model.database.DatabaseType;
import uk.ac.roslin.ensembl.config.EnsemblDBType;
import uk.ac.roslin.ensembl.config.EnsemblComparaDivision;
import uk.ac.roslin.ensembl.model.database.Registry;
import uk.ac.roslin.ensembl.model.database.SingleSpeciesCoreDatabase;

public class DBSpecies implements Species {

    protected String databaseStyleSpeciesName = "";
    protected String speciesBinomial = "";
    protected String commonName = "";
    protected String comparaName = "";
    protected String shortName = "";
    protected String mostRecentEnsemblSchemaVersion = "";
    protected String mostRecentDBRelease = "";
    protected String taxonomyID = "";
    protected String ensemblStablePrefix = "";
    protected TreeSet<String> aliases = new TreeSet<String>();
    protected Registry registry = null;
    private TreeSet<DBSingleSpeciesDatabase> databases = new TreeSet<DBSingleSpeciesDatabase>();
    private DBSingleSpeciesCoreDatabase currentCoreDatabase;
    private final Integer dBSpeciesID = 1;
    //this acts as the chromosome cache, hashed by version then by name
    TreeMap<String, TreeMap<String, DAChromosome>> chromosomes =
            new TreeMap<String, TreeMap<String, DAChromosome>>();
    private TreeSet<String> allChromosomesRetrieved = new TreeSet<String>();
    private TreeSet<String> versions = null;
    EnsemblDBType dbType;
    //default value
    EnsemblComparaDivision comparaDivision = EnsemblComparaDivision.MULTI;

    public DBSpecies() {
        dbType = EnsemblDBType.core;
    }

    public DBSpecies(DBSingleSpeciesCoreDatabase database) {
        this();
        this.registry = database.getRegistry();
        mostRecentEnsemblSchemaVersion = registry.getMostRecentEnsemblVersion();
        mostRecentDBRelease = registry.getMostRecentDBVersion();

        if (database.getDBVersion().equals(registry.getMostRecentDBVersion())) {
            this.currentCoreDatabase = database;
        }

        this.databases.add(database);

//        //from dbVersion 59 releases, the meta table holds the species.production_name,
//        //species.short_name and species.scientific_name meta_keys
//        //making this obsolete - but we will need to perform it in DBDatabaseDAO if
//        //it is a pre59 release
//        databaseStyleSpeciesName=this.currentCoreDatabase.getDbSpeciesName();
//        speciesBinomial = databaseStyleSpeciesName.replaceFirst( databaseStyleSpeciesName.substring(0, 1), (databaseStyleSpeciesName.substring(0,1)).toUpperCase() ).replace("_"," ");
//        aliases.add(speciesBinomial);

    }

    public TreeSet<String> getVersions() {
        if (versions != null) {
            return versions;
        } else {
            versions = new TreeSet<String>();
            for (DBDatabase db : this.getDatabasesByType(EnsemblDBType.core)) {
                versions.add(db.getDBVersion());
            }
            return versions;
        }
    }

    @Override
    public TreeSet<String> getAliases() {
        return aliases;
    }

    //will be overridden in DBCollectionSpecies
    @Override
    public void setProperty(HashMap row) {

        String key = (String) row.get("key");
        String value = (String) row.get("value");

        if (key.equals("species.stable_id_prefix")) {
            this.ensemblStablePrefix = value;
        } else if (key.equals("species.common_name")
                ||key.equals("species.ensembl_common_name")) {
            aliases.add(value);
            this.commonName = value;
        } else if (key.equals("species.short_name")) {
            aliases.add(value);
            this.shortName = value;
        } else if (key.equals("species.scientific_name")) {
            aliases.add(value);
            this.speciesBinomial = value;
        } else if (key.equals("species.production_name")) {
            aliases.add(value);
            this.databaseStyleSpeciesName = value;
        } else if (key.equals("species.taxonomy_id")) {
            aliases.add(value);
            this.taxonomyID = value;
        } else if (key.equals("species.alias")
           //     || key.equals("species.ensembl_common_name")
                || key.equals("species.ensembl_alias_name")) {
            aliases.add(value);
        }
    }

    @Override
    public CoreDatabase getCurrentCoreDatabase() {
        return currentCoreDatabase;
    }

    @Override
    public String getCommonName() {
        return commonName;
    }

    @Override
    public String getShortName() {
        return shortName;
    }

    public void setShortName(String name) {
        shortName = name;
    }

    @Override
    public String getDatabaseStyleName() {
        return databaseStyleSpeciesName;
    }

    @Override
    public TreeSet<? extends DBDatabase> getDatabases() {
        return databases;
    }

    @Override
    public void addDatabases(TreeSet<? extends Database> databases) {

        for (Database d : databases) {
            try {
                this.databases.add((DBSingleSpeciesDatabase) d);
            } catch (Exception e) {
            }
        }
    }

    @Override
    public void addDatabase(Database database) {

        try {
            this.databases.add((DBSingleSpeciesDatabase) database);
        } catch (Exception e) {
        }

    }

    @Override
    public String getEnsemblStablePrefix() {
        return ensemblStablePrefix;
    }

    @Override
    public String getSpeciesBinomial() {
        return speciesBinomial;
    }

    @Override
    public String getTaxonomyID() {
        return taxonomyID;
    }

    @Override
    public TreeSet<? extends DBDatabase> getDatabasesByType(DatabaseType type) {
        TreeSet<DBSingleSpeciesDatabase> out = new TreeSet<DBSingleSpeciesDatabase>();
        for (DBDatabase d : databases) {
            if (d.getType() == type) {
                out.add((DBSingleSpeciesDatabase) d);
            }
        }
        return out;
    }

    @Override
    public TreeSet<? extends DBDatabase> getDatabasesByVersion(String version) {
        TreeSet<DBSingleSpeciesDatabase> out = new TreeSet<DBSingleSpeciesDatabase>();
        for (DBDatabase d : databases) {
            if (d.getDBVersion().equals(version)) {
                out.add((DBSingleSpeciesDatabase) d);
            }
        }
        return out;
    }

    @Override
    public DBDatabase getDatabaseByTypeAndVersion(DatabaseType type, String version) {

        for (DBDatabase d : databases) {
            if (d.getDBVersion().equalsIgnoreCase(version) && d.getType() == type) {
                return d;
            }
        }
        return null;
    }

    public void setCommonName(String commonName) {
        this.commonName = commonName;
    }

    public void setEnsemblStablePrefix(String ensemblStablePrefix) {
        this.ensemblStablePrefix = ensemblStablePrefix;
    }

    public void setSpeciesBinomial(String speciesBinomial) {
        this.speciesBinomial = speciesBinomial;
    }

    public void setTaxonomyID(String taxonomyID) {
        this.taxonomyID = taxonomyID;
    }

    @Override
    public Integer getDBSpeciesID(String version) {
        return this.dBSpeciesID;
    }

    public Integer getDBSpeciesID() {
        return this.dBSpeciesID;
    }

    @Override
    public String getComparaName() {
        //for a single species Database the compara database name is the binomial
        //but for a collection species the productionname/comparaname/dbname
        return this.comparaName;
    }

    public void setComparaName(String comparaName) {
        //for a single species Database the compara database name is the binomial
        //but for a collection species the productionname/comparaname/dbname
        this.comparaName = comparaName;
    }

    @Override
    public TreeMap<String, DAChromosome> getChromosomes() throws DAOException {
        return getChromosomes(null);
    }

    @Override
    public TreeMap<String, DAChromosome> getChromosomes(String dbVersion) throws DAOException {

        if (dbVersion == null || dbVersion.isEmpty()) {
            dbVersion = mostRecentDBRelease;
        }

        if (!this.getVersions().contains(dbVersion)) {
            throw new DAOException("No version " + dbVersion + " for " + this.toString());
        }

        if (this.allChromosomesRetrieved.contains(dbVersion)) {
            return this.chromosomes.get(dbVersion);
        }

        List<DAChromosome> chrs = null;

        if (!this.chromosomes.containsKey(dbVersion)) {
            this.chromosomes.put(dbVersion, new TreeMap<String, DAChromosome>());
        }

        chrs = ((DBSingleSpeciesCoreDatabase) this.getDatabaseByTypeAndVersion(EnsemblDBType.core, dbVersion)).getChromosomes();

        this.allChromosomesRetrieved.add(dbVersion);

        for (DAChromosome c : chrs) {
            if (!chromosomes.get(dbVersion).containsKey(c.getChromosomeName())) {
                chromosomes.get(dbVersion).put(c.getChromosomeName(), c);
            }
        }

        return this.chromosomes.get(dbVersion);
    }

    @Override
    public DAChromosome getChromosomeByName(String name) throws DAOException {
        return this.getChromosomeByName(name, null);
    }

    @Override
    public DAChromosome getChromosomeByName(String name, String dbVersion) throws DAOException {
        DAChromosome chr = null;

        if (dbVersion == null || dbVersion.isEmpty()) {
            dbVersion = mostRecentDBRelease;
        }

        if (!this.getVersions().contains(dbVersion)) {
            throw new DAOException("No version " + dbVersion + " for " + this.toString());
        }

        if (!this.chromosomes.containsKey(dbVersion)) {

            this.chromosomes.put(dbVersion, new TreeMap<String, DAChromosome>());
        }

        if (this.chromosomes.get(dbVersion).containsKey(name)) {
            return this.chromosomes.get(dbVersion).get(name);
        } else {

            chr = ((DBSingleSpeciesCoreDatabase) this.getDatabaseByTypeAndVersion(EnsemblDBType.core, dbVersion)).getChromosomeByName(name);
        }

        if (chr != null) {
            chromosomes.get(dbVersion).put(name, chr);
        }

        return chr;
    }

    @Override
    public DAGene getGeneByID(Integer id) throws DAOException {
        return this.getGeneByID(id, null);
    }

    @Override
    public DAGene getGeneByID(Integer id, String dbVersion) throws DAOException {

        DAGene gene = null;

        if (dbVersion == null || dbVersion.isEmpty()) {
            dbVersion = this.mostRecentDBRelease;
        }

        try {
        gene = ((DBSingleSpeciesCoreDatabase) this.getDatabaseByTypeAndVersion(
                EnsemblDBType.core, dbVersion)).getCoreFactory().getGeneDAO().getGeneByID(id);
        } catch (Exception e) {
            if (e instanceof DAOException) {
                throw (DAOException) e;
            } else {
                throw new DAOException(e);
            }
        }

        return gene;
    }

    @Override
    public DAGene getGeneByStableID(String stableID) throws DAOException {
        return this.getGeneByStableID(stableID, null);
    }

    @Override
    public DAGene getGeneByStableID(String stableID, String dbVersion) throws DAOException {

        DAGene gene = null;

        if (dbVersion == null || dbVersion.isEmpty()) {
            dbVersion = this.mostRecentDBRelease;
        }

       try {
        gene = ((DBSingleSpeciesCoreDatabase) this.getDatabaseByTypeAndVersion(
                EnsemblDBType.core, dbVersion)).getCoreFactory().getGeneDAO().getGeneByStableID(stableID.trim());
            } catch (Exception e) {
            if (e instanceof DAOException) {
                throw (DAOException) e;
            } else {
                throw new DAOException(e);
            }
        }

        return gene;
    }

    @Override
    public TreeSet<String> getSchemaVersions() {
        TreeSet<String> out = new TreeSet<String>();

        for (Database d : this.getDatabasesByType(dbType)) {
            out.add(d.getSchemaVersion());
        }

        return out;
    }

    @Override
    public TreeSet<String> getDBVersions() {
        TreeSet<String> out = new TreeSet<String>();

        for (Database d : this.getDatabasesByType(dbType)) {
            out.add(d.getDBVersion());
        }

        return out;
    }

    public void setDatabaseStyleSpeciesName(String dbSpeciesName) {
        this.databaseStyleSpeciesName = dbSpeciesName;
    }

    @Override
    public String toString() {
        return ((this.commonName!= null && !this.commonName.isEmpty() ) ? this.commonName : this.shortName )
                + " (" + this.speciesBinomial + ")  ";
    }

    public DAChromosome getCachedChromosome(DAChromosome chr_in) {

        if (chr_in == null) {
            return null;
        }

        //test we have enough details on this chromosome to use the cache
        if (chr_in.getChromosomeName()== null
                || chr_in.getChromosomeName().isEmpty()
                || chr_in.getDaoFactory() == null
                || chr_in.getDaoFactory().getDBVersion()==null
                || chr_in.getDaoFactory().getDBVersion().isEmpty() ) {
            return chr_in;
        }

        //if the cach has this db version
        if (chromosomes.containsKey(chr_in.getDaoFactory().getDBVersion())) {

            //if the version hash has a chr of this name return the cached one
            if (chromosomes.get(chr_in.getDaoFactory().getDBVersion())
                    .containsKey(chr_in.getChromosomeName())) {
                return chromosomes.get(chr_in.getDaoFactory().getDBVersion())
                        .get(chr_in.getChromosomeName());
            }
            //otherwise add this chr to the cache and return it
            else {
                chromosomes.get(chr_in.getDaoFactory().getDBVersion())
                        .put(chr_in.getChromosomeName(), chr_in);
                return chr_in;
            }

        }
        //put create a hash for this version, and add this named chromosome to it, before returning it
        else {
            chromosomes.put(chr_in.getDaoFactory().getDBVersion(), new TreeMap<String, DAChromosome>());
            chromosomes.get(chr_in.getDaoFactory().getDBVersion())
                    .put(chr_in.getChromosomeName(), chr_in);
            return chr_in;
        }
    }

    public EnsemblComparaDivision getComparaDivision() {
        return comparaDivision;
    }

    public void setComparaDivision(String speciesDivision) {
        this.comparaDivision = EnsemblComparaDivision.getEnsemblComparaDivisionByMetaValue(speciesDivision);
    }

    @Override
    public String getAssembly() {
        return this.getAssembly(null);
    }

    @Override
    public String getAssembly(String dbVersion) {

        if (dbVersion == null || dbVersion.isEmpty()) {
            dbVersion = mostRecentDBRelease;
        }
        
      if (this.getDatabaseByTypeAndVersion(EnsemblDBType.core, dbVersion)!=null) {
          return ((SingleSpeciesCoreDatabase)
                  this.getDatabaseByTypeAndVersion(EnsemblDBType.core, dbVersion)).getAssembly();
      } else {
            return null;
      }
    }


}
