/**
 * Copyright (C) 2011 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */

package uk.ac.roslin.ensembl.dao.database.coreaccess;

import uk.ac.roslin.ensembl.exception.DAOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import org.apache.ibatis.session.SqlSession;
import org.apache.log4j.Logger;
import uk.ac.roslin.ensembl.datasourceaware.core.DAGene;
import uk.ac.roslin.ensembl.config.FeatureType;
import uk.ac.roslin.ensembl.model.Mapping;
import uk.ac.roslin.ensembl.datasourceaware.DAXRef;
import uk.ac.roslin.ensembl.datasourceaware.core.DAAssembledDNASequence;
import uk.ac.roslin.ensembl.datasourceaware.core.DAChromosome;
import uk.ac.roslin.ensembl.datasourceaware.core.DADNASequence;
import uk.ac.roslin.ensembl.mapper.handler.DBResultHandler;
import uk.ac.roslin.ensembl.mapper.query.FeatureQuery;
import uk.ac.roslin.ensembl.model.Coordinate;
import uk.ac.roslin.ensembl.model.core.CoreObject;
import uk.ac.roslin.ensembl.config.EnsemblCoordSystemType;
import uk.ac.roslin.ensembl.dao.coreaccess.ExonDAO;
import uk.ac.roslin.ensembl.dao.factory.DAOCollectionCoreFactory;
import uk.ac.roslin.ensembl.dao.factory.DAOCoreFactory;
import uk.ac.roslin.ensembl.dao.factory.DAOSingleSpeciesCoreFactory;
import uk.ac.roslin.ensembl.datasourceaware.core.DAExon;
import uk.ac.roslin.ensembl.datasourceaware.core.DATranscript;
import uk.ac.roslin.ensembl.mapper.core.ExonMapper;
import uk.ac.roslin.ensembl.mapper.core.TranscriptMapper;
import uk.ac.roslin.ensembl.model.MappingSet;
import uk.ac.roslin.ensembl.model.core.CoordinateSystem;
import uk.ac.roslin.ensembl.model.core.Gene;
import uk.ac.roslin.ensembl.model.core.Transcript;


public class DBExonDAO extends DBCoreObjectDAO implements ExonDAO {

    //if the meta table tells us that exon.build.level is 'toplevel'
    //we can grab the top ranked CS and use this for all genes
    //there may be other values than 'toplevel' - i dont know what :)
    //if the build level isn't annotated we have to get the possible CS levels
    //from meta_coord and use which ever one is approriate!
    CoordinateSystem exonBuildCS = null;
    final static Logger LOGGER = Logger.getLogger(DAExon.class);


    public DBExonDAO() {
        super();
    }

    public DBExonDAO(DAOSingleSpeciesCoreFactory factory) throws DAOException {
        super(factory);
        exonBuildCS = this.ssFactory.getDatabase().getBuildCoordSystem(FeatureType.exon.toString());
        //what if this is null or throws an exception?
        //if it throws an exception we are probably in some sort of bad factory environment
        if (exonBuildCS==null) {
            exonBuildCS=this.ssFactory.getDatabase().getTopLevelCoordSystem();
        }
    }

    public DBExonDAO(DAOCollectionCoreFactory factory) throws DAOException {
        super(factory);
        exonBuildCS= this.collFactory.getDatabase().getBuildCoordSystem(species, FeatureType.exon.toString());
        //what if this is null  or throw an exception?
        //if it throws an exception we are probably in some sort of bad factory environment
        if (exonBuildCS==null) {
            exonBuildCS=this.collFactory.getDatabase().getTopLevelCS(species);
        }
    }

    /**
     * Uses the stableid of an object to fill in missing data 
     * @param object
     * @throws DAOException
     */
    @Override
    public void reInitialize(CoreObject object) throws DAOException {

        if (object == null ||  !( object instanceof DAExon) ) {
            throw new DAOException("Object not a DAExon");
        }

        DAExon exon = (DAExon) object;

        //the DAO method requires a stableID
        if (exon.getStableID()==null || exon.getStableID().isEmpty() ) {
            return;
        }

        DAExon temp = this.getExonByStableID(exon.getStableID());


        exon.setId(temp.getId());
        exon.setCreationDate(temp.getCreationDate());
        exon.setModificationDate(temp.getModificationDate());
        exon.setDescription(temp.getDescription());

//        for (Mapping tempMapping :temp.getMappings()) {
//
//            Mapping tempRevMapping = tempMapping.getReverseMapping();
//            if (tempRevMapping != null) {
//                tempRevMapping.setTarget(transcript);
//            }
//            tempMapping.setSource(transcript);
//            transcript.addMapping(tempMapping);
//
//        }

    }

    /**
     * @param id
     * @return
     * @throws DAOException
     */
    @Override
    public DAExon getExonByID(Integer id) throws DAOException {
        
        if (id==null) {
            return null;
        }

        FeatureQuery q = new FeatureQuery();
        q.setFeatureID(id);
        if (!singleSpecies) {
            try {
                q.setSpeciesID(species.getDBSpeciesID(this.getFactory().getDBVersion()));
            } catch (Exception e) {
                throw new DAOException("No species ID context for this query!", e);
            }
        }
        return this.getExon(q);
    }

    /**
     * @param stableID
     * @return
     * @throws DAOException
     */
    @Override
    public DAExon getExonByStableID(String stableID) throws DAOException {
        if (stableID==null || stableID.isEmpty()  ) {
            return null;
        }

        FeatureQuery q = new FeatureQuery();
        if (!singleSpecies) {
            try {
                q.setSpeciesID(species.getDBSpeciesID(this.getFactory().getDBVersion()));
            } catch (Exception e) {
                throw new DAOException("No species ID context for this query!", e);
            }
        }
        q.setFeatureStableID(stableID.trim());
        return this.getExon(q);
    }

    @Override
    public List<DAExon> getExonsForTranscript(Transcript transcript) throws DAOException {
       if (transcript==null  || !(transcript instanceof DATranscript) || transcript.getId()==null ) {
            return null;
        }
       
       DATranscript daTranscript = (DATranscript) transcript;
              
        FeatureQuery q = new FeatureQuery();
        if (!singleSpecies) {
            try {
                q.setSpeciesID(species.getDBSpeciesID(this.getFactory().getDBVersion()));
            } catch (Exception e) {
                throw new DAOException("No species ID context for this query!", e);
            }
        }
       
        q.setTranscriptID(daTranscript.getId());
        
        return this.getExons(q, daTranscript);

    }

   //*******************************

    private DAExon getExon(FeatureQuery query) throws DAOException {
        DAExon out = null;
        ExonRowHandler handler = null;
        List<HashMap> result;

        SqlSession session = null;

        try {
            session = this.getFactory().getNewSqlSession();
            ExonMapper mapper = session.getMapper(ExonMapper.class);
            result = mapper.getExon(query);
        } catch (Exception e) {
            throw new DAOException("Failed to call getExon", e);
        } finally {
            if (session != null) {
                session.close();
            }
        }

        if (result != null && !result.isEmpty()) {

            handler = new ExonRowHandler(result);
            handler.handleResult();
            out = handler.getObjectResult();
        }

        return out;
    }

    private List<DAExon> getExons(FeatureQuery query, DATranscript transcript) throws DAOException {
        List<DAExon> out = null;
        ExonRowHandler handler = null;
        List<HashMap> result;

        SqlSession session = null;

        try {
            session = this.getFactory().getNewSqlSession();
            ExonMapper mapper = session.getMapper(ExonMapper.class);
            result = mapper.getExon(query);
        } catch (Exception e) {
            throw new DAOException("Failed to call getExons", e);
        } finally {
            if (session != null) {
                session.close();
            }
        }

        if (result != null && !result.isEmpty()) {

            handler = new ExonRowHandler(result,transcript);
            handler.handleResult();
            out = handler.getListResult();
        }

        return out;
    }

    //*********************************

    //as an inner class it has access to the factory etc of the enclosing class
    public class ExonRowHandler implements DBResultHandler {

        // the magic strings to be used as property keys for HashMap in Ibatis
        private final String exon = "exon";
        private final String target = "target";
        private final String coords = "coords";
        private final String coordSystemID = "csID";

        private DADNASequence parentSeq = null;
        private DAExon objectResult = null;
        private List<DAExon> listResult = new ArrayList<DAExon>();
        private List<HashMap> rawResults = null;
        private DATranscript daTranscript = null;

        public ExonRowHandler(List<HashMap> results, DATranscript t) {
            daTranscript = t;
            rawResults = results;
            try {
                MappingSet m = daTranscript.getTopLevelMappings();
                parentSeq = (DADNASequence) m.first().getTarget();
            } catch (DAOException ex) {
                    LOGGER.info("DAOException trying to retrieve the Chromosome for the Gene", ex);
            }

        }

        public ExonRowHandler(List<HashMap> results) {
            rawResults = results;
        }

        @Override
        public List<DAExon> getListResult() {
            return listResult;
        }

        @Override
        public DAExon getObjectResult() {
            return objectResult;
        }

        public void handleResult() throws DAOException {

            if (rawResults == null || rawResults.isEmpty()) {
                return;
            }

            for (HashMap map : rawResults) {
                handleRow(map);
            }

        }

        private void handleRow(HashMap result) throws DAOException {

            if (result == null || result.isEmpty()) {
                return;
            }

            DAExon rowExon = this.parseResult(result);


            if (rowExon != null) {
                this.objectResult = rowExon;
                this.listResult.add(rowExon);
            }

        }

        private DAExon parseResult(HashMap result) throws DAOException {

            if (result == null || result.isEmpty()) {
                return null;
            }

            DADNASequence pparentSeq = parentSeq;
            DAExon pexon = null;
            DADNASequence ptarget = null;
            Coordinate pcoords = null;
            Integer pcsID = null;

            pexon = (DAExon) result.get(this.exon);
            pcoords = (Coordinate) result.get(this.coords);
            ptarget = (DADNASequence) result.get(this.target);
            pcsID = (Integer) result.get(this.coordSystemID);

            if (pexon == null
                    || pcoords == null
                        || ptarget == null
                            ) {
                return null;
            }

            //if we havent got a target DNA for the Gene, we can make and use one
            //for the transcript, beware - may make >1 identically id'd sequence 
            //here tho cos not caching
            if (pparentSeq==null) {

                CoordinateSystem targetCS = null;

                if (ptarget.getDaoFactory() == null) {
                        ptarget.setDaoFactory(daoFactory);
                    }

                targetCS = ptarget.getCoordSystem();

                if (targetCS == null) {
                        if (singleSpecies) {
                            ptarget.setCoordSystem(ssFactory.getDatabase().getCSByID(pcsID));
                        } else {
                            ptarget.setCoordSystem(collFactory.getDatabase().getCSByID(species, pcsID));
                        }
                        targetCS = ptarget.getCoordSystem();
                    }


                 if (targetCS.isSequenceLevel()) {
                    ptarget.setCoordSystem(targetCS);
                 } else if (targetCS.getType().equals(EnsemblCoordSystemType.chromosome)) {
                    ptarget = new DAChromosome((DAOCoreFactory) daoFactory);
                    ptarget.setId(((DADNASequence) result.get(this.target)).getId());
                    ptarget.setSpecies(species);
                    ptarget.setCoordSystem(targetCS);

                    //loook to see if we have this sequence in cache
                    ptarget = species.getCachedChromosome((DAChromosome) ptarget);

                } else {
                    ptarget = new DAAssembledDNASequence((DAOCoreFactory) daoFactory);
                    ptarget.setId(((DADNASequence) result.get(this.target)).getId());
                    ptarget.setCoordSystem(targetCS);

                    //not got a cache of assembled sequences yet to check
                }


            } else {
                //if the target DNASequence isn't the sequence that the gene is on
                //throw this result away
                //ultimately will need to modify this to map from alternate CS/sequences
                if (!ptarget.getId().equals(pparentSeq.getId())) {
                    return null;                }
            }

            if (pexon.getDaoFactory() == null) {
                //pgene.setType(FeatureType.transcript);
                pexon.setDaoFactory(daoFactory);
            }


//            //check if the transcript object has somehow been returned from a cache and it
//            //has a mapping to a seq region with the same id as the parentSeq: if
//            //so, swap it for the parentSeq
//            for (Mapping m : pgene.getMappings()) {
//                if (m.getTarget().getId().equals(pparentSeq.getId())) {
//                    if (pparentSeq.getDaoFactory() == null) {
//                        pparentSeq.setDaoFactory(daoFactory);
//                    }
//                    if (pparentSeq.getCoordSystem() == null) {
//                        if (singleSpecies) {
//                            pparentSeq.setCoordSystem(ssFactory.getDatabase().getCSByID(pcsID));
//                        } else {
//                            pparentSeq.setCoordSystem(collFactory.getDatabase().getCSByID(species, pcsID));
//                        }
//                    }
//                    //do the switch
//                    m.setTarget(pparentSeq);
//                    m.setTargetCoordinates(pcoords);
//                    Mapping.addReverseMapping(m);
//                    return pgene;
//                }
//            }
//
//
            Mapping mapping = new Mapping();
            mapping.setSource(pexon);
            mapping.setTargetCoordinates(pcoords);
            mapping.setTarget(pparentSeq);

            if(pexon.addMapping(mapping)) {
                Mapping.addReverseMapping(mapping);
            }
            if (daTranscript!=null) {
                daTranscript.addExon(pexon);
                pexon.setTranscript(daTranscript);
            }
            return pexon;


        }

    
    }

}
