/**
 * Copyright (C) 2011 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */

package uk.ac.roslin.ensembl.dao.database.coreaccess;

import uk.ac.roslin.ensembl.exception.DAOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import org.apache.ibatis.session.SqlSession;
import org.apache.log4j.Logger;
import uk.ac.roslin.ensembl.datasourceaware.core.DAGene;
import uk.ac.roslin.ensembl.config.FeatureType;
import uk.ac.roslin.ensembl.model.Mapping;
import uk.ac.roslin.ensembl.datasourceaware.DAXRef;
import uk.ac.roslin.ensembl.datasourceaware.core.DAAssembledDNASequence;
import uk.ac.roslin.ensembl.datasourceaware.core.DAChromosome;
import uk.ac.roslin.ensembl.datasourceaware.core.DADNASequence;
import uk.ac.roslin.ensembl.mapper.handler.DBResultHandler;
import uk.ac.roslin.ensembl.mapper.query.FeatureQuery;
import uk.ac.roslin.ensembl.model.Coordinate;
import uk.ac.roslin.ensembl.model.core.CoreObject;
import uk.ac.roslin.ensembl.config.EnsemblCoordSystemType;
import uk.ac.roslin.ensembl.dao.coreaccess.TranscriptDAO;
import uk.ac.roslin.ensembl.dao.factory.DAOCollectionCoreFactory;
import uk.ac.roslin.ensembl.dao.factory.DAOCoreFactory;
import uk.ac.roslin.ensembl.dao.factory.DAOSingleSpeciesCoreFactory;
import uk.ac.roslin.ensembl.datasourceaware.core.DATranscript;
import uk.ac.roslin.ensembl.mapper.core.TranscriptMapper;
import uk.ac.roslin.ensembl.model.MappingSet;
import uk.ac.roslin.ensembl.model.core.CoordinateSystem;
import uk.ac.roslin.ensembl.model.core.Gene;


public class DBTranscriptDAO extends DBCoreObjectDAO implements TranscriptDAO {

    //if the meta table tells us that transcript.build.level is 'toplevel'
    //we can grab the top ranked CS and use this for all genes
    //there may be other values than 'toplevel' - i dont know what :)
    //if the build level isn't annotated we have to get the possible CS levels
    //from meta_coord and use which ever one is approriate!
    CoordinateSystem transcriptBuildCS = null;
    final static Logger LOGGER = Logger.getLogger(DBTranscriptDAO.class);


    public DBTranscriptDAO() {
        super();
    }

    public DBTranscriptDAO(DAOSingleSpeciesCoreFactory factory) throws DAOException {
        super(factory);

        //we want to initialize the GeneDAO so that it knows which COordinate system
        //it needs to quwery about genes

        transcriptBuildCS = this.ssFactory.getDatabase().getBuildCoordSystem(FeatureType.transcript.toString());
        //what if this is null or throws an exception?
        //if it throws an exception we are probably in some sort of bad factory environment
        if (transcriptBuildCS==null) {
            transcriptBuildCS=this.ssFactory.getDatabase().getTopLevelCoordSystem();
        }
    }

    public DBTranscriptDAO(DAOCollectionCoreFactory factory) throws DAOException {
        super(factory);
        transcriptBuildCS= this.collFactory.getDatabase().getBuildCoordSystem(species, FeatureType.transcript.toString());
        //what if this is null  or throw an exception?
        //if it throws an exception we are probably in some sort of bad factory environment
        if (transcriptBuildCS==null) {
            transcriptBuildCS=this.collFactory.getDatabase().getTopLevelCS(species);
        }
    }

    /**
     * Uses the stableid of an object to fill in missing data 
     * @param object
     * @throws DAOException
     */
    @Override
    public void reInitialize(CoreObject object) throws DAOException {

        if (object == null ||  !( object instanceof DATranscript) ) {
            throw new DAOException("Object not a DATranscript");
        }

        DATranscript transcript = (DATranscript) object;

        //the DAO method requires a stableID
        if (transcript.getStableID()==null || transcript.getStableID().isEmpty() ) {
            return;
        }

        DATranscript temp = this.getTranscriptByStableID(transcript.getStableID());

        transcript.setDisplayXRef(temp.getDisplayXRef());
        transcript.setBiotype(temp.getBiotype());
        transcript.setId(temp.getId());
        transcript.setCreationDate(temp.getCreationDate());
        transcript.setModificationDate(temp.getModificationDate());
        transcript.setDescription(temp.getDescription());

//        for (Mapping tempMapping :temp.getMappings()) {
//
//            Mapping tempRevMapping = tempMapping.getReverseMapping();
//            if (tempRevMapping != null) {
//                tempRevMapping.setTarget(transcript);
//            }
//            tempMapping.setSource(transcript);
//            transcript.addMapping(tempMapping);
//
//        }

    }

    /**
     * @param id
     * @return
     * @throws DAOException
     */
    @Override
    public DATranscript getTranscriptByID(Integer id) throws DAOException {
        
        if (id==null) {
            return null;
        }

        FeatureQuery q = new FeatureQuery();
        q.setFeatureID(id);
        if (!singleSpecies) {
            try {
                q.setSpeciesID(species.getDBSpeciesID(this.getFactory().getDBVersion()));
            } catch (Exception e) {
                throw new DAOException("No species ID context for this query!", e);
            }
        }
        return this.getTranscript(q);
    }

    /**
     * @param stableID
     * @return
     * @throws DAOException
     */
    @Override
    public DATranscript getTranscriptByStableID(String stableID) throws DAOException {
        if (stableID==null || stableID.isEmpty()  ) {
            return null;
        }

        FeatureQuery q = new FeatureQuery();
        if (!singleSpecies) {
            try {
                q.setSpeciesID(species.getDBSpeciesID(this.getFactory().getDBVersion()));
            } catch (Exception e) {
                throw new DAOException("No species ID context for this query!", e);
            }
        }
        q.setFeatureStableID(stableID.trim());
        return this.getTranscript(q);
    }

    @Override
    public List<DATranscript> getTranscriptsForGene(Gene gene) throws DAOException {
       if (gene==null  || !(gene instanceof DAGene) || gene.getId()==null ) {
            return null;
        }
       
       DAGene daGene = (DAGene) gene;
              
        FeatureQuery q = new FeatureQuery();
        if (!singleSpecies) {
            try {
                q.setSpeciesID(species.getDBSpeciesID(this.getFactory().getDBVersion()));
            } catch (Exception e) {
                throw new DAOException("No species ID context for this query!", e);
            }
        }
       
        q.setGeneID(daGene.getId());
        
        return this.getTranscripts(q, daGene);

    }

   //*******************************

    private DATranscript getTranscript(FeatureQuery query) throws DAOException {
        DATranscript out = null;
        TranscriptRowHandler handler = null;
        List<HashMap> result;

        SqlSession session = null;

        try {
            session = this.getFactory().getNewSqlSession();
            TranscriptMapper mapper = session.getMapper(TranscriptMapper.class);
            result = mapper.getTranscript(query);
        } catch (Exception e) {
            throw new DAOException("Failed to call getTranscript", e);
        } finally {
            if (session != null) {
                session.close();
            }
        }

        if (result != null && !result.isEmpty()) {

            handler = new TranscriptRowHandler(result);
            handler.handleResult();
            out = handler.getObjectResult();
        }

        return out;
    }

    private List<DATranscript> getTranscripts(FeatureQuery query, DAGene gene) throws DAOException {
        List<DATranscript> out = null;
        TranscriptRowHandler handler = null;
        List<HashMap> result;

        SqlSession session = null;

        try {
            session = this.getFactory().getNewSqlSession();
            TranscriptMapper mapper = session.getMapper(TranscriptMapper.class);
            result = mapper.getTranscript(query);
        } catch (Exception e) {
            throw new DAOException("Failed to call getTranscripts", e);
        } finally {
            if (session != null) {
                session.close();
            }
        }

        if (result != null && !result.isEmpty()) {

            handler = new TranscriptRowHandler(result,gene);
            handler.handleResult();
            out = handler.getListResult();
        }

        return out;
    }

    //*********************************

    //as an inner class it has access to the factory etc of the enclosing class
    public class TranscriptRowHandler implements DBResultHandler {

        // the magic strings to be used as property keys for HashMap in Ibatis
        private final String transcript = "transcript";
        private final String xref = "xref";
        private final String target = "target";
        private final String coords = "coords";
        private final String coordSystemID = "csID";

        private DADNASequence parentSeq = null;
        private DATranscript objectResult = null;
        private List<DATranscript> listResult = new ArrayList<DATranscript>();
        private List<HashMap> rawResults = null;
        private DAGene daGene = null;

        public TranscriptRowHandler(List<HashMap> results, DAGene g) {
            daGene = g;
            rawResults = results;
            try {
                MappingSet m = daGene.getTopLevelMappings();
                parentSeq = (DADNASequence) m.first().getTarget();
            } catch (DAOException ex) {
                    LOGGER.info("DAOException trying to retrieve the Chromosome for the Gene", ex);
            }

        }

        public TranscriptRowHandler(List<HashMap> results) {
            rawResults = results;
        }

        @Override
        public List<DATranscript> getListResult() {
            return listResult;
        }

        @Override
        public DATranscript getObjectResult() {
            return objectResult;
        }

        public void handleResult() throws DAOException {

            if (rawResults == null || rawResults.isEmpty()) {
                return;
            }

            for (HashMap map : rawResults) {
                handleRow(map);
            }

        }

        private void handleRow(HashMap result) throws DAOException {

            if (result == null || result.isEmpty()) {
                return;
            }

            DATranscript rowTranscript = this.parseResult(result);


            if (rowTranscript != null) {
                this.objectResult = rowTranscript;
                this.listResult.add(rowTranscript);
            }

        }

        private DATranscript parseResult(HashMap result) throws DAOException {

            if (result == null || result.isEmpty()) {
                return null;
            }

            DADNASequence pparentSeq = parentSeq;
            DATranscript ptranscript = null;
            DADNASequence ptarget = null;
            Coordinate pcoords = null;
            DAXRef pxref = null;
            Integer pcsID = null;

            ptranscript = (DATranscript) result.get(this.transcript);
            pcoords = (Coordinate) result.get(this.coords);
            ptarget = (DADNASequence) result.get(this.target);
            pxref = (DAXRef) result.get(this.xref);
            pcsID = (Integer) result.get(this.coordSystemID);

            if (ptranscript == null
                    || pcoords == null
                        || ptarget == null
                            ) {
                return null;
            }

            //if we havent got a target DNA for the Gene, we can make and use one
            //for the transcript, beware - may make >1 identically id'd sequence 
            //here tho cos not caching
            if (pparentSeq==null) {

                CoordinateSystem targetCS = null;

                if (ptarget.getDaoFactory() == null) {
                        ptarget.setDaoFactory(daoFactory);
                    }

                targetCS = ptarget.getCoordSystem();

                if (targetCS == null) {
                        if (singleSpecies) {
                            ptarget.setCoordSystem(ssFactory.getDatabase().getCSByID(pcsID));
                        } else {
                            ptarget.setCoordSystem(collFactory.getDatabase().getCSByID(species, pcsID));
                        }
                        targetCS = ptarget.getCoordSystem();
                    }


                 if (targetCS.isSequenceLevel()) {
                    ptarget.setCoordSystem(targetCS);
                 } else if (targetCS.getType().equals(EnsemblCoordSystemType.chromosome)) {
                    ptarget = new DAChromosome((DAOCoreFactory) daoFactory);
                    ptarget.setId(((DADNASequence) result.get(this.target)).getId());
                    ptarget.setSpecies(species);
                    ptarget.setCoordSystem(targetCS);

                    //loook to see if we have this sequence in cache
                    ptarget = species.getCachedChromosome((DAChromosome) ptarget);

                } else {
                    ptarget = new DAAssembledDNASequence((DAOCoreFactory) daoFactory);
                    ptarget.setId(((DADNASequence) result.get(this.target)).getId());
                    ptarget.setCoordSystem(targetCS);

                    //not got a cache of assembled sequences yet to check
                }


            } else {
                //if the target DNASequence isn't the sequence that the gene is on
                //throw this result away
                //ultimately will need to modify this to map from alternate CS/sequences
                if (!ptarget.getId().equals(pparentSeq.getId())) {
                    return null;                }
            }

            if (ptranscript.getDaoFactory() == null) {
                //pgene.setType(FeatureType.transcript);
                ptranscript.setDaoFactory(daoFactory);
            }

            if (ptranscript.getDisplayXRef() == null) {
                pxref = (DAXRef) result.get(this.xref);
                if (pxref != null) {
                    pxref.setDaoFactory(daoFactory);
                    ptranscript.setDisplayXRef(pxref);
                }
            }





//            //check if the transcript object has somehow been returned from a cache and it
//            //has a mapping to a seq region with the same id as the parentSeq: if
//            //so, swap it for the parentSeq
//            for (Mapping m : pgene.getMappings()) {
//                if (m.getTarget().getId().equals(pparentSeq.getId())) {
//                    if (pparentSeq.getDaoFactory() == null) {
//                        pparentSeq.setDaoFactory(daoFactory);
//                    }
//                    if (pparentSeq.getCoordSystem() == null) {
//                        if (singleSpecies) {
//                            pparentSeq.setCoordSystem(ssFactory.getDatabase().getCSByID(pcsID));
//                        } else {
//                            pparentSeq.setCoordSystem(collFactory.getDatabase().getCSByID(species, pcsID));
//                        }
//                    }
//                    //do the switch
//                    m.setTarget(pparentSeq);
//                    m.setTargetCoordinates(pcoords);
//                    Mapping.addReverseMapping(m);
//                    return pgene;
//                }
//            }
//
//
            Mapping mapping = new Mapping();
            mapping.setSource(ptranscript);
            mapping.setTargetCoordinates(pcoords);
            mapping.setTarget(pparentSeq);

            if(ptranscript.addMapping(mapping)) {
                Mapping.addReverseMapping(mapping);
            }
            if (daGene!=null) {
                if (daGene.getCanonicalTranscriptID()!= null
                        && ptranscript.getId() != null
                            && daGene.getCanonicalTranscriptID().compareTo(ptranscript.getId())==0) {
                    ptranscript.setCanonical(true);
                }
                daGene.addTranscript(ptranscript);
                ptranscript.setGene(daGene); 
            }
            return ptranscript;


        }

    
    }

}
