/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */
package uk.ac.roslin.ensembl.datasourceaware.core;

import java.util.Iterator;
import java.util.LinkedList;
import org.biojava3.core.sequence.Strand;
import uk.ac.roslin.ensembl.model.Mapping;
import uk.ac.roslin.ensembl.model.MappingSet;
import uk.ac.roslin.ensembl.exception.DAOException;
import uk.ac.roslin.ensembl.exception.RangeException;
import uk.ac.roslin.ensembl.model.Coordinate;
import uk.ac.roslin.ensembl.model.Mapping;
import uk.ac.roslin.ensembl.model.core.AssembledDNASequence;
import uk.ac.roslin.ensembl.model.core.Assembly;

/**
 *
 * @author paterson
 */
public class DAAssembly implements Assembly {

    DAAssembledDNASequence parent = null;
    protected MappingSet componentMappings = null;
    protected MappingSet stitchedComponentMappings = null;
    protected Integer assemblyStart = null;
    protected Integer assemblyStop = null;

    public static DAAssembly getAssembly(DAAssembledDNASequence parent) {
        return new DAAssembly(parent);
    }

    public static DAAssembly getAssembly(DAAssembledDNASequence parent, Integer start, Integer stop) {
        return new DAAssembly(parent, start, stop);
    }

    public DAAssembly() {
    }

    private DAAssembly(DAAssembledDNASequence parent) {
        this.parent = parent;
    }

    private DAAssembly(DAAssembledDNASequence parent, Integer start, Integer stop) {
        this.parent = parent;
        this.assemblyStart = start;
        this.assemblyStop = stop;
    }

    public DAAssembledDNASequence getParent() {
        return parent;
    }

    public void setParent(AssembledDNASequence parent) {
        this.parent = (DAAssembledDNASequence) parent;
    }

    public Integer getAssemblyStart() {
        if (assemblyStart == null) {
            try {
                assemblyStart = (parent.getBioBegin() != null) ? parent.getBioBegin() : 1;
            } catch (Exception e) {
            }
        }
        if (assemblyStart == null) {
            assemblyStart = 1;
        }
        return assemblyStart;
    }

    public void setAssemblyStart(Integer assemblyStart) {
        this.assemblyStart = assemblyStart;
    }

    public Integer getAssemblyStop() {
        if (assemblyStop != null || parent==null) {
            return assemblyStop;
        }
        assemblyStop = parent.getBioEnd() ;
        return assemblyStop;
    }

    public void setAssemblyStop(Integer assemblyStop) {
        this.assemblyStop = assemblyStop;
    }

    private void setMappings() throws DAOException {
        componentMappings = new MappingSet();
        try {
            componentMappings = (MappingSet) parent.getDaoFactory().getAssemblyDAO().getComponentMappingsByStartStop(parent, this.getAssemblyStart(), this.getAssemblyStop());
        } catch (Exception e) {
                throw new DAOException("Failed to create the stitched mappings for a DAAssembly", e);
        }
     }

    public MappingSet getMappings() throws DAOException {
        if (this.componentMappings == null) {
            this.setMappings();
        }
        return this.componentMappings;
    }

    public MappingSet getStitchedMappings() throws DAOException {
        if (this.stitchedComponentMappings == null) {
        this.stitchedComponentMappings = this.stitchComponents(
                        this.getAssemblyStart(), this.getAssemblyStop(), this.getMappings());
        }
        return this.stitchedComponentMappings;
    }

    public MappingSet getStitchedMappings(Integer start, Integer stop) throws DAOException {
        if (this.stitchedComponentMappings == null) {
            this.stitchedComponentMappings = this.stitchComponents(this.getAssemblyStart(), this.getAssemblyStop(), this.getMappings());
        }

        MappingSet out = new MappingSet();

        for (Mapping mapping : this.getStitchedMappings()) {


            if (mapping.getSourceCoordinates().getStart() > stop) {
                break;
            } else if (mapping.getSourceCoordinates().getEnd() < start) {
                continue;
            } else {
                Integer upstream = mapping.getSourceCoordinates().getStart() - start;
                Integer downstream = mapping.getSourceCoordinates().getEnd() - stop;

                if (upstream >= 0 && downstream <= 0) {
                    out.add((Mapping) mapping);
                    continue;
                }
                Mapping newMapping = new Mapping();
                newMapping.setSource(mapping.getSource());
                newMapping.setSourceCoordinates(mapping.getSourceCoordinates());
                newMapping.setTarget(mapping.getTarget());
                newMapping.setTargetCoordinates(mapping.getTargetCoordinates());

                if (upstream < 0) {
                    newMapping.getTargetCoordinates().setStart(mapping.getTargetCoordinates().getStart() - upstream);
                    newMapping.getSourceCoordinates().setStart(start);
                }
                if (downstream > 0) {
                    newMapping.getTargetCoordinates().setEnd(mapping.getTargetCoordinates().getEnd() - downstream);
                    newMapping.getSourceCoordinates().setEnd(stop);
                }
                out.add(newMapping);
                continue;
            }
        }

        return out;
    }

    private MappingSet stitchComponents(Integer _start, Integer _end, MappingSet mappings) throws DAOException {

        MappingSet out = new MappingSet();

        boolean reverse = false;
        Integer start;
        Integer end;

        if (_start==null || _end ==null) {
            throw new DAOException("invalid assembly coordinates");
        }

        if (_start > _end) {
            reverse = true;
            start = _end;
            end = _start;
        } else {
            start = _start;
            end = _end;
        }

        Integer currentPosition = start;


        for (Mapping mapping : mappings) {

            Integer sourceStart = mapping.getSourceCoordinates().getStart();
            Integer sourceEnd = mapping.getSourceCoordinates().getEnd();
            DADNASequence source = (DADNASequence) mapping.getSource();
            DADNASequence target = (DADNASequence) mapping.getTarget();
            Integer distanceLeft = end - currentPosition;

            //skip ones completely before the desired start
            if (sourceStart > end || sourceEnd < currentPosition) {
                continue;
            } else if (sourceStart <= currentPosition) {

                //we overlap the start
                Integer overlap = currentPosition - sourceStart;
                Mapping m = new Mapping();
                m.setSource(source);
                m.setTarget(target);

                //if we also overlap the end
                if (sourceEnd > end) {
                    m.setTargetCoordinates(
                            mapping.getTargetCoordinates().getStart() + overlap,
                            mapping.getTargetCoordinates().getStart() + overlap + distanceLeft,
                            mapping.getTargetCoordinates().getStrand());
                    m.setSourceCoordinates(currentPosition, end, Coordinate.Strand.FORWARD_STRAND);
                    out.add(m);
                    //we're done
                    currentPosition = end + 1;
                    break;
                } else {
                    m.setTargetCoordinates(
                            mapping.getTargetCoordinates().getStart() + overlap,
                            mapping.getTargetCoordinates().getEnd(),
                            mapping.getTargetCoordinates().getStrand());
                    m.setSourceCoordinates(currentPosition, sourceEnd, Coordinate.Strand.FORWARD_STRAND);
                    out.add(m);
                    if (sourceEnd == end) {
                        //we're done
                        currentPosition = end + 1;
                        break;
                    } else {
                        currentPosition = sourceEnd + 1;
                        continue;
                    }
                }



            } else if (currentPosition < sourceStart) {

                //plug a gap
                Integer gapLength = sourceStart - currentPosition;
                Mapping m = new Mapping();
                m.setSource(source);
                m.setTarget(GapSequence.makeGap(gapLength));
                m.setTargetCoordinates(1, gapLength, Coordinate.Strand.FORWARD_STRAND);
                m.setSourceCoordinates(currentPosition, currentPosition + gapLength - 1, Coordinate.Strand.FORWARD_STRAND);
                out.add(m);
                currentPosition = currentPosition + gapLength;


                //if we are within bounds
                if (end >= sourceEnd) {
                    Mapping m1 = new Mapping();
                    m1.setSource(source);
                    m1.setTarget(target);
                    m1.setSourceCoordinates(currentPosition, sourceEnd, Coordinate.Strand.FORWARD_STRAND);
                    m1.setTargetCoordinates(mapping.getTargetCoordinates());
                    out.add(m1);
                    if (end == sourceEnd) {
                        // we're done
                        currentPosition = end + 1;
                        break;
                    } else {
                        currentPosition = sourceEnd + 1;
                        continue;
                    }
                } //if we run past the desired end
                else {
                    Mapping m2 = new Mapping();
                    m2.setSource(source);
                    m2.setTarget(target);
                    m2.setSourceCoordinates(currentPosition, end, Coordinate.Strand.FORWARD_STRAND);
                    m2.setTargetCoordinates(1, end - sourceStart + 1, mapping.getTargetCoordinates().getStrand());
                    out.add(m2);
                    //we're done
                    currentPosition = end + 1;
                    break;
                }

            }

        }

        if (currentPosition <= end) {
            //plug a terminal gap
            Integer gapLength = end - currentPosition + 1;
            Mapping m = new Mapping();
            m.setSource(parent);
            m.setTarget(GapSequence.makeGap(gapLength));
            m.setTargetCoordinates(1, gapLength, Coordinate.Strand.FORWARD_STRAND);
            m.setSourceCoordinates(currentPosition, end, Coordinate.Strand.FORWARD_STRAND);
            out.add(m);

        }

        return out;
    }

    public String getSequenceAsString(Integer start, Integer stop, Strand strand) throws RangeException, DAOException {

        StringBuffer sb = new StringBuffer();
        String out = "";

        //range of sequence I want
        Integer begin = start;
        Integer end = stop;

        //make sure range is not greater than the assembly extent
        if (end == null || this.getAssemblyStop() ==null || end > this.getAssemblyStop()) {
            // end = this.getAssemblyStop();
            throw new RangeException("Requested range greater than assembly range.");
        }
        if (begin == null || this.getAssemblyStart() ==null || begin < this.getAssemblyStart()) {
            //begin = this.getAssemblyStart();
            throw new RangeException("Requested range greater than assembly range.");
        }

        Integer nextStart = start;

        for (Mapping mapping : this.getStitchedMappings()) {

            if (nextStart > stop) {
                //we're done
                break;
            }

            //////////////////////////////////////
            if (mapping.getSourceCoordinates().getStart() > stop) {
                //we're past the end - so quit
                break;
            } else if (mapping.getSourceCoordinates().getEnd() < nextStart) {
                //we're not yet at the position we want so skip
                continue;
            } else {

                //this mapping is at least partially in range

                Integer desiredOutputStart = mapping.getTargetCoordinates().getStart() + (nextStart - mapping.getSourceCoordinates().getStart());
                Integer desiredOutputEnd = mapping.getTargetCoordinates().getStart() + (stop - mapping.getSourceCoordinates().getStart());


                if (desiredOutputEnd > mapping.getTargetCoordinates().getEnd()) {
                    desiredOutputEnd = mapping.getTargetCoordinates().getEnd();
                } else {
                    //this will complete the sequence
                }

                nextStart = nextStart + desiredOutputEnd - desiredOutputStart + 1;

                DADNASequence seq = (DADNASequence) mapping.getTarget();
                if (mapping.getTargetCoordinates().getStrand().equals(uk.ac.roslin.ensembl.model.Coordinate.Strand.FORWARD_STRAND)) {
                    sb.append(((DADNASequence) ((Mapping) mapping).getTarget()).getSequenceAsString(
                            desiredOutputStart, desiredOutputEnd, null));
                } else {
                    sb.append(((DADNASequence) ((Mapping) mapping).getTarget()).getReverseComplementSequenceAsString(seq.getLength() - desiredOutputEnd + 1, seq.getLength() - desiredOutputStart + 1));
                }
            }
        }
        return sb.toString();
    }

    public String getReverseComplementSequenceAsString(Integer start, Integer stop) throws RangeException, DAOException {

        StringBuffer outerBuffer = new StringBuffer();
        LinkedList<StringBuffer> buffers = new LinkedList<StringBuffer>();

        String out = "";

        //range of sequence I want
        Integer begin = start;
        Integer end = stop;

        //make sure range is not greater than the assembly extent
        if (end == null || end > this.getAssemblyStop()) {
            // end = this.getAssemblyStop();
            throw new RangeException("Requested range greater than assembly range.");
        }
        if (begin == null || begin < this.getAssemblyStart()) {
            //begin = this.getAssemblyStart();
            throw new RangeException("Requested range greater than assembly range.");
        }

        Integer nextStart = start;

        for (Mapping mapping : this.getStitchedMappings()) {

            if (nextStart > stop) {
                //we're done
                break;
            }

            if (mapping.getSourceCoordinates().getStart() > stop) {
                //we're past the end - so quit
                break;
            } else if (mapping.getSourceCoordinates().getEnd() < nextStart) {
                //we're not yet at the position we want so skip
                continue;
            } else {

                //this mapping is at least partially in range
                Integer desiredOutputStart = mapping.getTargetCoordinates().getStart() + (nextStart - mapping.getSourceCoordinates().getStart());
                Integer desiredOutputEnd = mapping.getTargetCoordinates().getStart() + (stop - mapping.getSourceCoordinates().getStart());

                if (desiredOutputEnd > mapping.getTargetCoordinates().getEnd()) {
                    desiredOutputEnd = mapping.getTargetCoordinates().getEnd();
                } else {
                    //this will complete the sequence
                }
                nextStart = nextStart + desiredOutputEnd - desiredOutputStart + 1;
                StringBuffer innerBuffer = new StringBuffer();
                DADNASequence seq = (DADNASequence) mapping.getTarget();

                if (mapping.getTargetCoordinates().getStrand().equals(uk.ac.roslin.ensembl.model.Coordinate.Strand.FORWARD_STRAND)) {

                    innerBuffer.append(((DADNASequence) ((Mapping) mapping).getTarget()).getReverseComplementSequenceAsString(
                            desiredOutputStart,desiredOutputEnd));
                            
                } else {
                    innerBuffer.append(((DADNASequence) ((Mapping) mapping).getTarget()).getSequenceAsString(
                            //desiredOutputStart, desiredOutputEnd, null));
                            seq.getLength() - desiredOutputEnd + 1, seq.getLength() - desiredOutputStart + 1));
                }
                buffers.add(innerBuffer);
            }
        }

        Iterator<StringBuffer> it = buffers.descendingIterator();

        while (it.hasNext()) {
            outerBuffer.append(it.next());
        }




        return outerBuffer.toString();
    }
}
