/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */
package uk.ac.roslin.ensembl.datasourceaware.core;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import org.apache.log4j.Logger;
import uk.ac.roslin.ensembl.config.EnsemblCoordSystemType;
import uk.ac.roslin.ensembl.config.EnsemblDBType;
import uk.ac.roslin.ensembl.config.FeatureType;
import uk.ac.roslin.ensembl.model.MappingSet;
import uk.ac.roslin.ensembl.dao.factory.DAOCoreFactory;
import uk.ac.roslin.ensembl.datasourceaware.DAXRef;
import uk.ac.roslin.ensembl.exception.DAOException;
import uk.ac.roslin.ensembl.model.Mapping;
import uk.ac.roslin.ensembl.model.ObjectType;
import uk.ac.roslin.ensembl.model.core.Chromosome;
import uk.ac.roslin.ensembl.model.core.CoordinateSystem;
import uk.ac.roslin.ensembl.model.core.Feature;
import uk.ac.roslin.ensembl.model.database.CollectionCoreDatabase;
import uk.ac.roslin.ensembl.model.database.SingleSpeciesCoreDatabase;

/**
 *
 * @author tpaterso
 */
public abstract class DAFeature extends DACoreObject implements Feature {

    protected MappingSet mappings = new MappingSet();
    protected HashMap<ObjectType, MappingSet> objectTypeMappings = new HashMap<ObjectType, MappingSet>();
    protected Set<ObjectType> mappedObjectTypes = new HashSet<ObjectType>();
    protected Date creationDate = null;
    protected Date modificationDate = null;
    protected String description = null;
    protected DAXRef displayXRef = null;
    protected List<DAXRef> xrefs = new ArrayList<DAXRef>();
    protected String displayName = null;
    private Status status = Status.UNKNOWN;
    protected Boolean current = null;

    final static Logger LOGGER = Logger.getLogger(DAFeature.class);

    public DAFeature() {
        super();
    }

    public DAFeature(DAOCoreFactory factory) {
        super(factory);
    }

    public MappingSet getMappings() {
        //we haven't got the lazy load working here yet
        return this.mappings;
    }

    //Should Be Private?? and call by named object type methods ??
    public MappingSet getMappings(ObjectType targetType) {

        if (this.objectTypeMappings.containsKey(targetType)) {
            return this.objectTypeMappings.get(targetType);
        } else if (this.isObjectTypeMapped(targetType)) {
            this.objectTypeMappings.put(targetType, new MappingSet());
            return this.objectTypeMappings.get(targetType);
        } else {
            try {
                //we haven't got the lazy load working here yet
                this.reinitialize();
            } catch (DAOException ex) {
                LOGGER.warn("failed to reinitialize a feature by accessing database", ex);
            }
            return this.objectTypeMappings.get(targetType);
        }
    }

    public Boolean addMapping(Mapping mapping) {

//        //check we haven't already added a mapping for an object with this id!
//        if (mapping.getTarget() != null) {
//
//            for (Mapping m : this.mappings) {
//                if (m.getTarget() != null && m.getTarget().getId().equals(mapping.getTarget().getId())) {
//                    return;
//                }
//            }
//        }

        //if we fail to add the mapping this will be false
        if (this.mappings.add((Mapping) mapping)) {

            ObjectType t = mapping.getTargetType();

            if (t != null) {
                if (!this.objectTypeMappings.containsKey(t)) {
                    this.objectTypeMappings.put(t, new MappingSet());
                }

                this.objectTypeMappings.get(t).add((Mapping) mapping);
            }
            return true;
        }

        return false;

    }

    public void addMappedObjectType(ObjectType mappedType) {
        mappedObjectTypes.add(mappedType);
    }

    public Boolean isObjectTypeMapped(ObjectType mappedType) {
        return mappedObjectTypes.contains(mappedType);
    }

    // can there be more than one of these?
    public MappingSet getTopLevelMappings() throws DAOException {

        CoordinateSystem topCS = null;
        MappingSet out = new MappingSet();

        if (this.getDaoFactory()==null) {
            throw new DAOException("No DAOFactory Set on this DAFeature");
        }

        try {
                if (this.getDaoFactory().getDBType().equals(EnsemblDBType.core)) {

                    topCS = ((SingleSpeciesCoreDatabase) this.getDaoFactory().getDatabase()).getTopLevelCoordSystem();

                } else if (this.getDaoFactory().getDBType().equals(EnsemblDBType.collection_core)) {
                    topCS = ((CollectionCoreDatabase) this.getDaoFactory().getDatabase()).getTopLevelCS(this.getSpecies());

                }
            } catch (DAOException dex) {
                LOGGER.warn("Failed to get the top level coordinate System for a Feature "
                        +this.getClass().getSimpleName()+ " "+this.getId(), dex);
                throw dex;
            } catch (Exception ex) {
                LOGGER.warn("Failed to get the top level coordinate System for a Feature "
                        +this.getClass().getSimpleName()+ " "+this.getId(), ex);
                throw new DAOException("Failed to retrieve top level CoordinateSystem for this CoreDatabase", ex);
            }

        if (topCS != null) {

            out = this.getMappings(topCS.getType());
            this.addMappedObjectType(topCS.getType());
        } else {
            return null;
        }


        return out;
    }

    /**
     * Utility method to pull back a single mapping of this Feature on a Given chromosome.
     * This should be the mapping stored at initialisation.
     * If the Feature has implemented a resinitialize method this may be called.
     * @param chr
     * @return a single Mapping
     */
    public Mapping getChromosomeMapping(Chromosome chr) {
        
        if (chr==null) {
            return null;
        }
        MappingSet s = this.getMappings(EnsemblCoordSystemType.chromosome);
        for (Mapping m : s) {
            if (m.getTarget()==chr) {
                return m;
            }
        }
        return null;
    }
    
    public MappingSet getAnnotationLevelMappings() throws DAOException  {

        if (this.getDaoFactory()==null) {
            throw new DAOException("No DAOFactory Set on this DAFeature");
        }

        Set<? extends CoordinateSystem> annotCS = null;

        try {
            if (this.getDaoFactory().getDBType().equals(EnsemblDBType.core)) {

                annotCS = ((SingleSpeciesCoreDatabase) this.getDaoFactory().getDatabase()).getCSForFeature(this.getType());

            } else if (this.getDaoFactory().getDBType().equals(EnsemblDBType.collection_core)) {
                annotCS = ((CollectionCoreDatabase) this.getDaoFactory().getDatabase())
                        .getCSForFeature(this.getSpecies(),this.getType());
            }
        } catch (DAOException dex) {
                LOGGER.warn("Failed to get the annotation level coordinate System for a Feature "
                        +this.getClass().getSimpleName()+ " "+this.getId(), dex);
                throw dex;
            } catch (Exception ex) {
                LOGGER.warn("Failed to get the annotation level coordinate System for a Feature "
                        +this.getClass().getSimpleName()+ " "+this.getId(), ex);
                throw new DAOException("Failed to retrieve annotation level CoordinateSystem for this CoreDatabase", ex);
            }

        if (annotCS==null || annotCS.isEmpty()) {
            LOGGER.warn("Failed to get the annotation level coordinate System for a Feature "
                    +this.getClass().getSimpleName()+ " "+this.getId());
            return null;
        }

        MappingSet out = new MappingSet();

        for (CoordinateSystem cs : annotCS ) {

            MappingSet t = this.getMappings(cs.getType());
            this.addMappedObjectType(cs.getType());
            if (t != null) {
               out.addAll(t);
            }

        }

        return out;
    }

    public MappingSet getBuildLevelMappings() throws DAOException{

        if (this.getDaoFactory()==null) {
            throw new DAOException("No DAOFactory Set on this DAFeature");
        }

        MappingSet out = new MappingSet();
        CoordinateSystem buildCS = null;
        
        try {
            if (this.getDaoFactory().getDBType().equals(EnsemblDBType.core)) {

                buildCS = ((SingleSpeciesCoreDatabase) this.getDaoFactory().getDatabase()).getBuildCoordSystem(this.getType().toString());

            } else if (this.getDaoFactory().getDBType().equals(EnsemblDBType.collection_core)) {
                buildCS = ((CollectionCoreDatabase) this.getDaoFactory().getDatabase()).getBuildCoordSystem(this.getSpecies(),this.getType().toString() );

            }
        } catch (DAOException dex) {
                LOGGER.warn("Failed to get the build level coordinate system for this database "
                        +this.getClass().getSimpleName()+ " "+this.getId(), dex);
                throw dex;
            } catch (Exception ex) {
                LOGGER.warn("Failed to get the build level coordinate system for this database "
                        +this.getClass().getSimpleName()+ " "+this.getId(), ex);
                throw new DAOException("Failed to retrieve build level CoordinateSystem for this CoreDatabase", ex);
            }

        if (buildCS != null) {

            out = this.getMappings(buildCS.getType());
            this.addMappedObjectType(buildCS.getType());
        } else {
            return null;
        }

        return out;
    }

    public void clearAllMappings() {
            mappings.clear();
            mappedObjectTypes.clear();
            objectTypeMappings.clear();
    }

    public String getDescription()  {
        return description;
    }

    public void setDescription(String description) {
        this.description = description;
    }

    public Date getCreationDate() {
        return creationDate;
    }

    public void setCreationDate(Date creationDate) {
        this.creationDate = creationDate;
    }

    public Date getModificationDate() {
        return modificationDate;
    }

    public void setModificationDate(Date modificationDate) {
        this.modificationDate = modificationDate;
    }

    public DAXRef getDisplayXRef() {
        return displayXRef;
    }

    public void setDisplayXRef(DAXRef xref) {
        this.displayXRef = xref;
        if (xref!=null) {
            this.xrefs.add(xref);
        }
    }

    public List<DAXRef> getXRefs() {
        return xrefs;
    }

    public void addXRefs(List<DAXRef> xrefs) {
        this.xrefs.addAll(xrefs);
    }

    abstract void reinitialize() throws DAOException ;

    public static enum Status {
        KNOWN,
        NOVEL,
        PUTATIVE,
        PREDICTED,
        KNOWN_BY_PROJECTION,
        UNKNOWN;
    }

    public String getStatus() {
        return this.status.toString();
    }

    public void setStatus(String status) {
        try {
            this.status = Status.valueOf(status);
        } catch (IllegalArgumentException e) {
            this.status = Status.UNKNOWN;
        }
    }

    public String getDisplayName() {
        return displayName ;
    }

    public void setDisplayName(String displayName) {
        this.displayName = displayName;
    }

    public Boolean isCurrent() {
        return current;
    }

    public void setCurrent(Boolean current) {
        this.current = current;
    }
 
}
