/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */

package uk.ac.roslin.ensembl.datasourceaware.core;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.TreeMap;
import java.util.TreeSet;
import org.apache.log4j.Logger;
import uk.ac.roslin.ensembl.config.EnsemblComparaDivision;
import uk.ac.roslin.ensembl.dao.factory.DAOCoreFactory;
import uk.ac.roslin.ensembl.exception.DAOException;
import uk.ac.roslin.ensembl.model.ObjectType;
import uk.ac.roslin.ensembl.model.core.Analysis;
import uk.ac.roslin.ensembl.model.core.Gene;
import uk.ac.roslin.ensembl.model.core.Transcript;
import uk.ac.roslin.ensembl.config.FeatureType;
import uk.ac.roslin.ensembl.datasourceaware.compara.DAHomologyPairRelationship;
import uk.ac.roslin.ensembl.model.core.Species;

public class DAGene extends DAFeature implements Gene {


    private String biotype = null;
    private DAAnalysis analysis = null;
    private Integer canonicalTranscriptID = null;
    private String stableID = null;
    private TreeMap<Integer, DATranscript> transcriptHash = null;
    private HashMap<EnsemblComparaDivision, List<DAHomologyPairRelationship>> homologies =
            new HashMap<EnsemblComparaDivision, List<DAHomologyPairRelationship>>();
    private TreeSet<EnsemblComparaDivision> homologySearchesPerformed = new TreeSet<EnsemblComparaDivision>();

    final static Logger LOGGER = Logger.getLogger(DAGene.class);

    public DAGene() {
        super();
    }

    public DAGene(DAOCoreFactory factory) {
        super(factory);
     }

    @Override
    public ObjectType getType() {
        return FeatureType.gene;
    }

    public DAAnalysis getAnalysis() {
        return this.analysis;
    }

    public void setAnalysis(Analysis analysis) {
        this.analysis = (DAAnalysis) analysis;
    }

    public String getBiotype() {
        return this.biotype;
    }

    public void setBiotype(String biotype) {
        this.biotype = biotype;
    }

    public Collection<DATranscript> getTranscripts() {
        if (transcriptHash == null) {

            try {
                transcriptHash = new TreeMap<Integer, DATranscript>();
                return (Collection<DATranscript>) this.getDaoFactory().getTranscriptDAO().getTranscriptsForGene(this);
            } catch (Exception e) {
                transcriptHash = new TreeMap<Integer, DATranscript>();
                LOGGER.info("Threw DAOException on tryiong to populate transcripts for a Gene", e);
            }

        }
        return this.transcriptHash.values();
    }

    public void addTranscript(Transcript transcript) {
        if (transcriptHash==null) {
            transcriptHash = new TreeMap<Integer, DATranscript>();
        }
        try {
            DATranscript t = (DATranscript) transcript;
            this.transcriptHash.put(t.getId(), t);
        } catch (Exception e) {
            LOGGER.info("failed to add atranscript to the Gene", e);
        }
    }

    public DATranscript getCanonicalTranscript() {
        if (this.canonicalTranscriptID==null || this.canonicalTranscriptID == 0) {
            return null;
        }
        this.getTranscripts();   
        if (this.transcriptHash.containsKey(this.canonicalTranscriptID)) {
            return this.transcriptHash.get(this.canonicalTranscriptID);
        } else {
            return null;
        }
       
    }

    public String getStableID() {
        return stableID;
    }

    public void setStableID(String stableID) {
        this.stableID = stableID;
    }

    public Integer getCanonicalTranscriptID() {
        return canonicalTranscriptID;
    }

    public void setCanonicalTranscriptID(Integer canonicalTranscriptID) {
        this.canonicalTranscriptID = canonicalTranscriptID;
    }

    public DATranslation getCanonicalTranslation() {
        if (this.getCanonicalTranscript() != null) {
            return this.getCanonicalTranscript().getTranslation();
        } else {
            return null;
        }
    }
    
    /**
     * Returns all Homologies (as a List of HomologyPairRelationships) in the default
     * ComparaDivision for this Gene. Implemented by 
     * calling getHomologies(EnsemblComparaDivision comparaDivision) with
     * the default ComparaDivision for this Gene. 
     * @return
     */
    public List<DAHomologyPairRelationship> getHomologies() {
        return this.getHomologies(this.getComparaDivision());
    }

    /**
     * Returns all Homologies (as a List of HomologyPairRelationships) in the specified
     * ComparaDivision for this Gene.  
     * @param comparaDivision
     * @return
     */
    public List<DAHomologyPairRelationship> getHomologies(EnsemblComparaDivision comparaDivision) {

        List<DAHomologyPairRelationship> out = null;

        if (comparaDivision==null) {
            return out;
        }

        if (!this.homologySearchesPerformed.contains(comparaDivision)) {

            if (this.getStableID() == null || this.getStableID().isEmpty()) {
                return out;
            }

            try {
                out = (List<DAHomologyPairRelationship>) this.getComparaFactory(comparaDivision).getHomologyDAO().getHomologiesForGene(this);
                this.homologySearchesPerformed.add(comparaDivision);
                //this would over write any existing... maybe ok!
                this.homologies.put(comparaDivision, out);

//                if (this.homologies.get(comparaDivision) == null) {
//                    this.homologies.put(comparaDivision, new ArrayList<DAHomologyPairRelationship>());
//                }
//                this.homologies.get(comparaDivision).addAll(out);

            } catch (Exception ex) {
                LOGGER.info("failed to get Homologies", ex);
            }

        }

        if (this.homologies.containsKey(comparaDivision)) {
            return this.homologies.get(comparaDivision);
        } else {
            return new ArrayList<DAHomologyPairRelationship>();
        }
    }


    /**
     * Returns Homologies (as a List of HomologyPairRelationships) in the default
     * ComparaDivision for this Gene for the specified Species. Implemented by 
     * calling getHomologies(EnsemblComparaDivision comparaDivision, Species sp) with
     * the default ComparaDivision for this Gene. 
     * @param sp
     * @return
     */
    public List<DAHomologyPairRelationship> getHomologies(Species sp) {
        return this.getHomologies(this.getComparaDivision(), sp);
    }

    /**
     * Returns Homologies (as a List of HomologyPairRelationships) in the specified
     * ComparaDivision for the specified Species.
     * Cos of the lack of indices it is much faster to implement using the query
     * to get  all homologues than restrict to a single gdb.name in the sql [unless
     * you do it by a specific subselect]. So to get homologues for a single species do
     * the whole query and do the filterring in the Java. This allows us to store
     * all homologies at once too.
     * @param comparaDivision
     * @param sp
     * @return
     */
    public List<DAHomologyPairRelationship> getHomologies(EnsemblComparaDivision comparaDivision, Species sp) {

        //make sure basic fetch is done
        this.getHomologies(comparaDivision);
        
        List<DAHomologyPairRelationship> out = new ArrayList<DAHomologyPairRelationship>();

        if (homologies.get(comparaDivision)!=null && !homologies.get(comparaDivision).isEmpty()) {
            for (DAHomologyPairRelationship r : homologies.get(comparaDivision)) {
                if (r.getTarget().getSpecies()!=null
                        && r.getTarget().getSpecies().equals(sp)) {
                    out.add(r);
                }
            }
        }
        return out;
    }


    public void addHomology(EnsemblComparaDivision division, DAHomologyPairRelationship homology) {

        if (this.homologies.get(division)==null) {
            this.homologies.put(division, new ArrayList<DAHomologyPairRelationship>());
        }
        this.homologies.get(division).add(homology);

    }

    public boolean setSearchedHomologies (EnsemblComparaDivision comparaDivision) {
        return this.homologySearchesPerformed.add(comparaDivision);
    }
    
    @Override
    void reinitialize() throws DAOException
    {
        try {
            this.getDaoFactory().getGeneDAO().reInitialize(this);
        } catch (DAOException dex) {
            throw dex;
        } catch (Exception ex) {
            throw new DAOException("Failed to reinitialize a gene from the Database", ex);
        }
    }

    @Override
    public String getDisplayName() {
        return (displayName!=null) ? displayName : stableID ;
    }

    /**
     * Utility method to get all of the current DAHomologyPairRelationships for this Gene
     *
     * @return List<DAHomologyPair BinaryRelationship>
     */
    public List<DAHomologyPairRelationship> getHomologiesWithoutLazyLoad() {

        List<DAHomologyPairRelationship> out = new ArrayList<DAHomologyPairRelationship>();

        for (EnsemblComparaDivision div: homologies.keySet()) {
                out.addAll(homologies.get(div));
        }
        return out;
    }
}
