/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */

package uk.ac.roslin.ensembl.model;

import java.io.Serializable;
import java.util.TreeSet;

/**
 *
 * @author paterson
 */
public class Coordinate implements Comparable<Coordinate>, Serializable {

    Integer start = null;
    Integer end = null;
    Strand strand = null;
    private Integer strandInt = null;


    public Coordinate() {

    }

    public Coordinate(Integer start, Integer end) {

        this(start, end, Strand.strand(0));
    }

    public Coordinate(Integer start, Integer end, Integer strand) {

        this(start, end, Strand.strand(strand));
    }

    public Coordinate(Integer start, Integer end, Coordinate.Strand strand) {

        if (start!= null && end != null) {
            if (start<=end) {
                this.setStart(start);
                this.setEnd(end);
            } else {
                this.setStart(end);
                this.setEnd(start);
            }
        }
        this.setStrand(strand); 
    }

    @Override
    public String toString() {
        String out = "";
        out += this.start+" - "+this.end+" ("
                + (( this.strand != null) ? this.strand.toString() : "UNSPECIFIED_STRAND") +")";
        return out;
    }

    public static enum Strand implements Serializable  {

        FORWARD_STRAND,
        REVERSE_STRAND;

        public static Strand strand(int i) {
            if (i == 1) {
                return FORWARD_STRAND;
            } else if (i == -1) {
                return REVERSE_STRAND;
            } else {
                return null;
            }
        }
        public static Strand strand(Integer i) {
            if (i == 1) {
                return FORWARD_STRAND;
            } else if (i == -1) {
                return REVERSE_STRAND;
            } else {
                return null;
            }
        }
    }


    public Integer getStart() {
        return start;
    }

    public Integer getEnd() {
        return end;
    }

    public Strand getStrand() {
        return strand;
    }

    public void setStart(Integer b) {
        start = b;
    }

    public void setEnd(Integer e) {
            end = e;
        }

    public void setStrand(Strand s) {
            strand = s;
    }

    public void setStrandInt(Integer s) {
            strandInt = s;
            strand = Coordinate.Strand.strand(s);
    }

    public Integer getStrandInt() {
        return strandInt;
    }

//    public Integer getPlusStrandStart() {
//        if (strand == Strand.FORWARD_STRAND) {
//            return start;
//        } else  if (strand == Strand.REVERSE_STRAND) {
//            return end;
//        } else return null;
//    }
//
//    public Integer getPlusStrandEnd() {
//        if (strand == Strand.FORWARD_STRAND) {
//            return end;
//        } else  if (strand == Strand.REVERSE_STRAND) {
//            return start;
//        } else return null;
//    }

    //this can be used for ordering by the plus strand start, or the plus strand
    //end if the starts are identical

    //nb careful with Integer comparisons

    //nb 0 != identity , because it doesnt consider the strand, just the order

    public int compareTo(Coordinate o) {

        //dont compare if any nulls
        if (this.start==null || this.end==null
                || o.getStart()==null ||o.getEnd()==null) {
            return 0;
        }


        if (!this.start.equals(o.getStart())) {

            if (this.start.compareTo(o.getStart()) < 0) {
                return -1;
            } else {
                return 1;
            } 

        } else {
            //put the longest one first
            //this has consequences for getting the end coordinate of a set
            if (this.end.compareTo(o.getEnd()) < 0) {
                return 1;
            } else  if (this.end.compareTo(o.getEnd()) > 0) {
                return -1;
            } else {
                //the coordinates are identical
                return 0;
            }
        }
}

    public  Boolean liesWithinCoordinate(Coordinate test) {
        Boolean out = false;

        if (test.getStart()==null || test.getEnd()==null
                || this.start == null || this.end==null) {
            return out;
        }

        if (this.start>=test.getStart() && this.end<=test.getEnd()) {
            out = true;
        }

        return out;
    }

    public Boolean overlaps(Coordinate test) {

        boolean out = false;

        if (this.start==null || this.end==null || test.getEnd()==null || test.getStart()==null) {
            out = false;
//        } else if ( (this.start>=test.getStart() && this.start<=test.getEnd())
//                || (this.end>=test.getStart() && this.end<=test.getEnd() )) {
        } else if (
                (this.liesWithinCoordinate(test)) || (test.liesWithinCoordinate(this))
                || (this.start <= test.getStart() && this.end >= test.getStart())
                || (this.end >= test.getEnd() && this.start <= test.getEnd() )) {
            out = true;
        }

        return out;
    }

    public Coordinate getOverlap(Coordinate test) {

        Coordinate out = null;
        Integer start = test.getStart();
        Integer end = test.getEnd();


        if (this.start==null || this.end==null || test.getEnd()==null || test.getStart()==null
                || !this.overlaps(test)) {
            out = null;
        } else {
                if (this.start>=test.getStart() && this.start<=test.getEnd() ) {
                    start = this.start;
                }
                if (this.end>=test.getStart() && this.end<=test.getEnd() )
                {
                    end = this.end;
                }
                out = new Coordinate(start, end, 1);
        }

        return out;
    }



//    public static Integer getStart(TreeSet<Coordinate> testSet) {
//        Integer out = null;
//
//        if ( testSet != null && !testSet.isEmpty()) {
//            out = testSet.first().getStart();
//        }
//
//        return out;
//    }
//
//    public static Integer getEnd(TreeSet<Coordinate> testSet) {
//        Integer out = null;
//
//        if ( testSet != null && !testSet.isEmpty()) {
//
//            out = testSet.last().getEnd();
//
//            for (Coordinate c : testSet) {
//             if (c.getEnd()>out) {
//                 out = c.getEnd();
//             }
//            }
//        }
//         return out;
//    }
//
//    public static Coordinate getRange(TreeSet<Coordinate> testSet) {
//         Coordinate out = null;
//         if ( testSet != null && !testSet.isEmpty()) {
//
//             out = new Coordinate(Coordinate.getStart(testSet),Coordinate.getEnd(testSet) ,1);
//
//         }
//
//         return null;
//     }



//
//    /**
//     * tests whether this Coordinate represents a region fully covered in the
//     * parameter Set of Coordinates - without any gaps
//     * @param testSet TreeSet<Coordinate>
//     * @return Boolean
//     */
//    public  Boolean liesWithinCoordinateSetWithoutGaps(TreeSet<Coordinate> testSet) {
//        Boolean out = false;
//
//        Integer rangeBegin = Coordinate.getStart(testSet);
//        Integer rangeEnd = Coordinate.getEnd(testSet);
//
//        if (this.start<rangeBegin || this.end>rangeEnd ) {
//            out = false;
//        } else if (this.getGaps(testSet).isEmpty()) {
//            out = true;
//        }
//
//        return out;
//    }

//    /**
//     *
//     * @param testSet TreeSet<Coordinate>
//     * @return TreeSet<Coordinate>
//     */
//    public  TreeSet<Coordinate> getGaps(TreeSet<Coordinate> testSet) {
//
//        Integer rangeBegin = Coordinate.getStart(testSet);
//        Integer rangeEnd = Coordinate.getEnd(testSet);
//
//        TreeSet<Coordinate> gaps = new TreeSet<Coordinate>();
//        TreeSet<Coordinate> temp = Coordinate.getCoordinateGaps(testSet);
//
//
//        if (this.start<rangeBegin) {
//            Coordinate newC = new Coordinate(this.getStart(), rangeBegin-1, 1);
//            gaps.add(newC);
//        }
//
//        if (this.end>rangeEnd) {
//            Coordinate newC2 = new Coordinate(rangeEnd+1, this.end, 1);
//            gaps.add(newC2);
//        }
//
//
//
//        for (Coordinate cd : temp) {
//
//            //this gap completely contains the desired range
//            if (this.liesWithinCoordinate(cd)) {
//                Coordinate newC = new Coordinate(this.getStart(), this.getEnd(), 1);
//                gaps.add(newC);
//                break;
//            }
//
//            // this gap lies completely within the range
//            else if (cd.liesWithinCoordinate(this)) {
//                gaps.add(cd);
//            }
//
//            //this gap overlaps the beginning of the range
//            else if (cd.getStart()<=this.getStart() && cd.getEnd()>=this.getStart()) {
//                Coordinate newC = new Coordinate(this.getStart(), cd.getEnd(), 1);
//                gaps.add(newC);
//            }
//
//            //this gap overlies the end of the range
//            else if (cd.getEnd()>=this.getEnd() && cd.getStart()<=this.getEnd()) {
//                Coordinate newC = new Coordinate(cd.getStart(), this.getEnd(), 1);
//                gaps.add(newC);
//            }
//
//        }
//
//        return gaps;
//    }

//    public static TreeSet<Coordinate> getCoordinateGaps(TreeSet<Coordinate> testSet){
//        TreeSet<Coordinate> gaps = new TreeSet<Coordinate>();
//
//        Integer rangeBegin = Coordinate.getStart(testSet);
//        Integer rangeEnd = Coordinate.getEnd(testSet);
//        Integer currentPosition = Coordinate.getStart(testSet);
//
//        for (Coordinate c : testSet) {
//
//            Integer s = c.getStart();
//            Integer e = c.getEnd();
//
//            if (s>currentPosition) {
//
//                Coordinate newC = new Coordinate(currentPosition, s-1, 1);
//                gaps.add(newC);
//            }
//
//            if (e>=rangeEnd) {
//                break;
//            }
//
//
//            currentPosition = e+1;
//
//        }
//
//        return gaps;
//    }

}
