/**
 * Copyright (C) 2010 The Roslin Institute <contact andy.law@roslin.ed.ac.uk>
 *
 * This file is part of the Ensembl Java API demonstration project developed by the
 * Bioinformatics Group at The Roslin Institute, The Royal (Dick) School of
 * Veterinary Studies, University of Edinburgh.
 *
 * This is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License (version 3) as published by
 * the Free Software Foundation.
 *
 * This software is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * in this software distribution. If not, see <http://www.gnu.org/licenses/gpl-3.0.html/>.
 */
package uk.ac.roslin.ensembl.model.compara;

import java.io.ObjectStreamException;
import java.util.Collection;
import java.util.HashMap;
import uk.ac.roslin.ensembl.model.relationship.RelationshipType;

public class HomologyType extends RelationshipType {

    private static HashMap<String, HomologyType> typeHash;

    /**
     * A private constructor for making HomologyType objects. Should only be
     * accessed from within the class - and then only the first time that
     * the class is accessed when the list of acceptable types is constructed.
     * @param typeName The name of this HomologyType object
     */
    private HomologyType(String typeName) {
        super(typeName);
    }
    
    /**
     * two homologous objects derived by speciation
     * Ensembl pipeline annotates as ortholog_one2one
     * see http://Nov2010.archive.ensembl.org/info/docs/compara/homology_method.html
     */
    public static final HomologyType ORTHOLOG_ONE2ONE = HomologyType.makeHomologyType("ortholog_one2one");
    /**
     * two homologous objects derived by speciation
     * Ensembl pipeline annotates as ortholog_one2many
     * see http://Nov2010.archive.ensembl.org/info/docs/compara/homology_method.html
     */
    public static final HomologyType ORTHOLOG_ONE2MANY = HomologyType.makeHomologyType("ortholog_one2many");
    /**
     * two homologous objects derived by speciation
     * Ensembl pipeline annotates as ortholog_many2many
     * see http://Nov2010.archive.ensembl.org/info/docs/compara/homology_method.html
     */
    public static final HomologyType ORTHOLOG_MANY2MANY = HomologyType.makeHomologyType("ortholog_many2many");
    /**
     * two homologous objects derived by speciation
     * Ensembl pipeline annotates as apparent_ortholog_one2one
     * see http://Nov2010.archive.ensembl.org/info/docs/compara/homology_method.html
     */
    public static final HomologyType APPARENT_ORTHOLOG_ONE2ONE = HomologyType.makeHomologyType("apparent_ortholog_one2one");
    /**
     * two homologous objects derived by speciation
     * Ensembl pipeline annotates as possible_ortholog
     * see http://Nov2010.archive.ensembl.org/info/docs/compara/homology_method.html
     */
    public static final HomologyType POSSIBLE_ORTHOLOG = HomologyType.makeHomologyType("possible_ortholog");

    /**
     * two homologous objects derived by duplication
     * Ensembl pipeline annotates as within_species_paralog
     * see http://Nov2010.archive.ensembl.org/info/docs/compara/homology_method.html
     */
    public static final HomologyType WITHIN_SPECIES_PARALOG = HomologyType.makeHomologyType("within_species_paralog");
    /**
     * two homologous objects derived by duplication
     * Ensembl pipeline annotates as between_species_paralog
     * see http://Nov2010.archive.ensembl.org/info/docs/compara/homology_method.html
     */
    public static final HomologyType BETWEEN_SPECIES_PARALOG = HomologyType.makeHomologyType("between_species_paralog");
    /**
     * two homologous objects derived by duplication
     * Ensembl pipeline annotates as other_paralog
     * see http://Nov2010.archive.ensembl.org/info/docs/compara/homology_method.html
     */
    public static final HomologyType OTHER_PARALOG = HomologyType.makeHomologyType("other_paralog");
    /**
     * two homologous objects derived by duplication
     * Ensembl pipeline annotates as contiguous_gene_split
     * see http://Nov2010.archive.ensembl.org/info/docs/compara/homology_method.html
     */
    public static final HomologyType CONTIGUOUS_GENE_SPLIT = HomologyType.makeHomologyType("contiguous_gene_split");
    /**
     * two homologous objects derived by duplication
     * Ensembl pipeline annotates as putative_gene_split
     * see http://Nov2010.archive.ensembl.org/info/docs/compara/homology_method.html
     */
    public static final HomologyType PUTATIVE_GENE_SPLIT = HomologyType.makeHomologyType("putative_gene_split");

    /**
     * Ensure Singleton class
     */
    private Object readResolve() throws ObjectStreamException {
        return this;
    }

    /**
     * A private accessor to make a HomologyType object for a given string name.
     * This implements a singleton pattern against each name.
     * @param typeName The name of the HomologyType object to be generated
     * @return HomologyType
     */
    private static final HomologyType makeHomologyType(String typeName) {

        if (HomologyType.typeHash == null) {
            HomologyType.typeHash = new HashMap<String, HomologyType>();
        }

        HomologyType ht = getHomologyTypeByName(typeName);

        if (ht == null) {
            ht = new HomologyType(typeName);
            HomologyType.typeHash.put(typeName, ht);
        }
        return ht;
    }

    /**
     * Checks to see if the supplied type is the name of a valid HomologyType
     * and returns the corresponding HomologyType object if it is.
     * @param typeName
     * @return HomologyType - the HomologyType object corresponding to the name supplied
     */
    public static final HomologyType getHomologyTypeByName(String typeName) {
        return typeHash.get(typeName);
    }

    //remember to "override" static methods from the parent!
    public static HomologyType getRelationshipType(String value) {
        return typeHash.get(value);
    }

    //remember to "override" static methods from the parent!
    public static Collection<HomologyType> getAllTypes() {
        return typeHash.values();
    }
}
